/* decodeISUP.cpp
 * Includes code copied from the Ethereal source file packet-isup.c
 * Modifications made on 3/08/2004.
 * Copyright  2004 Verso Technologies, Inc. 
 */

/* packet-isup.c
 * Routines for ISUP dissection
 * Copyright 2001, Martina Obermeier <martina.obermeier@icn.siemens.de>
 *
 * $Id: packet-isup.c,v 1.22 2003/04/22 13:47:37 tuexen Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include "value_string.h"
#include <iostream>


char *data;

const char*
match_strval(unsigned int val, const value_string *vs) {
  int i = 0;

  while (vs[i].strptr) {
    if (vs[i].value == val)
      return(vs[i].strptr);
    i++;
  }

  return(NULL);
}


char* tvb_bytes_to_str(int offset, int len)
{
        static char str[254];
        int i=0;
        for (i = 0; i < (len*2); i++)
                str[i]=data[offset*2+i];
        str[i]=0;
        return &str[0];
}


const char* val_to_str(unsigned int val, const value_string *vs, const char *fmt) {
  const char *ret;
  static char  str[3][64];
  static char *cur;

  if (fmt == NULL) 
	  fmt = "Spare";
  ret = match_strval(val, vs);
  if (ret != NULL)
    return ret;
  if (cur == &str[0][0]) {
    cur = &str[1][0];
  } else if (cur == &str[1][0]) {
    cur = &str[2][0];
  } else {
    cur = &str[0][0];
  }
  sprintf(cur, fmt, val);
  return cur;
}


static const value_string dec_to_bin_value[] = 
{
  { 0,"0000"},
  { 1,"0001"},
  { 2,"0010"},
  { 3,"0011"},
  { 4,"0100"},
  { 5,"0101"},
  { 6,"0110"},
  { 7,"0111"},
  { 8,"1000"},
  { 9,"1001"},
  { 10,"1010"},
  { 11,"1011"},
  { 12,"1100"},
  { 13,"1101"},
  { 14,"1110"},
  { 15,"1111"}
};

char* tvb_bytes_to_BinaryStr(int offset, int len)
{
  static char str[1024];
  int i=0;
  int i1st4bits, i2d4bits;
  
  str[0]='\0';

  for (i=0; i<len;i++)
  {
	sscanf( &data[2*(offset+i)], "%1x", &i1st4bits);
    sscanf( &data[2*(offset+i)+1], "%1x", &i2d4bits);

	sprintf(&str[0], "%s %3d : %s %s", &str[0], i*8, val_to_str(i1st4bits,dec_to_bin_value,"error"), val_to_str(i2d4bits,dec_to_bin_value,"error"));
    if (i < (len-1))
		sprintf(&str[0], "%s \n%22s", &str[0]," ");
  }
  return &str[0];
}

const char* tf_to_str(unsigned int val, const true_false_string tfs) {
  if (val)
  {
	  return tfs.true_string;
  } else
  {
	  return tfs.false_string;
  };
}


/* Add a text-only node to the proto_tree */
void
proto_tree_add_text(int tab, const char *text, const char *format)
{
	switch (tab)
	{
		case 0:
			break;
		case 1:
            printf("%3s"," ");
			break;
		case 2:
			printf("%6s"," ");
			break;
		case 3:
			printf("%8s"," ");
			break;
		case 4:
			printf("%9s"," ");
			break;
	}
	printf(text, format);
	printf("\n");
}

void
proto_tree_add_text_i(int tab, const char *text, int value)
{
	switch (tab)
	{
		case 0:
			break;
		case 1:
            printf("%3s"," ");
			break;
		case 2:
			printf("%6s"," ");
			break;
		case 3:
			printf("%8s"," ");
			break;
		case 4:
			printf("%9s"," ");
			break;
	}
	printf(text, value);
	printf("\n");

}

void
proto_tree_add_text_u(int tab, const char *text, unsigned long value)
{
	switch (tab)
	{
		case 0:
			break;
		case 1:
            printf("%3s"," ");
			break;
		case 2:
			printf("%6s"," ");
			break;
		case 3:
			printf("%7s"," ");
			break;
		case 4:
			printf("%9s"," ");
			break;
	}
	printf(text, value);
	printf("\n");
}


unsigned char
tvb_get_guint8(int offset)
{
	int   i;
	sscanf( &data[2*offset], "%2x", &i );
	return i;
}

unsigned int
tvb_get_guint16(int offset)
{
	unsigned int   i;
	sscanf( &data[2*offset], "%4x", &i );
	return i;
}

unsigned int
tvb_get_guint16i(int offset)
{
	unsigned int   i, tmp;
	sscanf( &data[2*offset], "%2x", &i );
	sscanf( &data[2*(offset+1)], "%2x", &tmp );
	return (i+tmp*256);
}

unsigned long
tvb_get_guint24(int offset)
{
	unsigned long   i;
	sscanf( &data[2*offset], "%6x", &i );
	return i;
}


#include "decodeISDN.cpp"
// --------------------------------------------------------------------------------------------------------------
#define MTP3_ISUP_SERVICE_INDICATOR     5
#define ASCII_NUMBER_DELTA              0x30
#define ASCII_LETTER_DELTA              0x37

/* Definition of protocol field values und lengths */

/* Definition of Message Types */
#define MESSAGE_TYPE_INITIAL_ADDR       1
#define MESSAGE_TYPE_SUBSEQ_ADDR        2
#define MESSAGE_TYPE_INFO_REQ           3
#define MESSAGE_TYPE_INFO               4
#define MESSAGE_TYPE_CONTINUITY         5
#define MESSAGE_TYPE_ADDR_CMPL          6
#define MESSAGE_TYPE_CONNECT            7
#define MESSAGE_TYPE_FORW_TRANS         8
#define MESSAGE_TYPE_ANSWER             9
#define MESSAGE_TYPE_RELEASE           12
#define MESSAGE_TYPE_SUSPEND           13
#define MESSAGE_TYPE_RESUME            14
#define MESSAGE_TYPE_REL_CMPL          16
#define MESSAGE_TYPE_CONT_CHECK_REQ    17
#define MESSAGE_TYPE_RESET_CIRCUIT     18
#define MESSAGE_TYPE_BLOCKING          19
#define MESSAGE_TYPE_UNBLOCKING        20
#define MESSAGE_TYPE_BLOCK_ACK         21
#define MESSAGE_TYPE_UNBLOCK_ACK       22
#define MESSAGE_TYPE_CIRC_GRP_RST      23
#define MESSAGE_TYPE_CIRC_GRP_BLCK     24
#define MESSAGE_TYPE_CIRC_GRP_UNBL     25
#define MESSAGE_TYPE_CIRC_GRP_BL_ACK   26
#define MESSAGE_TYPE_CIRC_GRP_UNBL_ACK 27
#define MESSAGE_TYPE_FACILITY_REQ      31
#define MESSAGE_TYPE_FACILITY_ACC      32
#define MESSAGE_TYPE_FACILITY_REJ      33
#define MESSAGE_TYPE_LOOP_BACK_ACK     36
#define MESSAGE_TYPE_PASS_ALONG        40
#define MESSAGE_TYPE_CIRC_GRP_RST_ACK  41
#define MESSAGE_TYPE_CIRC_GRP_QRY      42
#define MESSAGE_TYPE_CIRC_GRP_QRY_RSP  43
#define MESSAGE_TYPE_CALL_PROGRSS      44
#define MESSAGE_TYPE_USER2USER_INFO    45
#define MESSAGE_TYPE_UNEQUIPPED_CIC    46
#define MESSAGE_TYPE_CONFUSION         47
#define MESSAGE_TYPE_OVERLOAD          48
#define MESSAGE_TYPE_CHARGE_INFO       49
#define MESSAGE_TYPE_NETW_RESRC_MGMT   50
#define MESSAGE_TYPE_FACILITY          51
#define MESSAGE_TYPE_USER_PART_TEST    52
#define MESSAGE_TYPE_USER_PART_AVAIL   53
#define MESSAGE_TYPE_IDENT_REQ         54
#define MESSAGE_TYPE_IDENT_RSP         55
#define MESSAGE_TYPE_SEGMENTATION      56
#define MESSAGE_TYPE_LOOP_PREVENTION   64

static const value_string isup_message_type_value[] = {
  { MESSAGE_TYPE_INITIAL_ADDR,          "Initial address"},
  { MESSAGE_TYPE_SUBSEQ_ADDR,           "Subsequent address"},
  { MESSAGE_TYPE_INFO_REQ,              "Information request (national use)"},
  { MESSAGE_TYPE_INFO,                  "Information (national use)"},
  { MESSAGE_TYPE_CONTINUITY,            "Continuity"},
  { MESSAGE_TYPE_ADDR_CMPL,             "Address complete"},
  { MESSAGE_TYPE_CONNECT,               "Connect"},
  { MESSAGE_TYPE_FORW_TRANS,            "Forward transfer"},
  { MESSAGE_TYPE_ANSWER,                "Answer"},
  { MESSAGE_TYPE_RELEASE,               "Release"},
  { MESSAGE_TYPE_SUSPEND,               "Suspend"},
  { MESSAGE_TYPE_RESUME,                "Resume"},
  { MESSAGE_TYPE_REL_CMPL,              "Release complete"},
  { MESSAGE_TYPE_CONT_CHECK_REQ,        "Continuity check request"},
  { MESSAGE_TYPE_RESET_CIRCUIT,         "Reset Circuit"},
  { MESSAGE_TYPE_BLOCKING,              "Blocking"},
  { MESSAGE_TYPE_UNBLOCKING,            "Unblocking"},
  { MESSAGE_TYPE_BLOCK_ACK,             "Blocking acknowledgement"},
  { MESSAGE_TYPE_UNBLOCK_ACK,           "Unblocking acknowledgment"},
  { MESSAGE_TYPE_CIRC_GRP_RST,          "Circuit group reset"},
  { MESSAGE_TYPE_CIRC_GRP_BLCK,         "Circuit group blocking"},
  { MESSAGE_TYPE_CIRC_GRP_UNBL,         "Circuit group unblocking"},
  { MESSAGE_TYPE_CIRC_GRP_BL_ACK,       "Circuit group blocking acknowledgement"},
  { MESSAGE_TYPE_CIRC_GRP_UNBL_ACK,     "Circuit group unblocking acknowledgement"},
  { MESSAGE_TYPE_FACILITY_REQ,          "Facility request"},
  { MESSAGE_TYPE_FACILITY_ACC,          "Facility accepted"},
  { MESSAGE_TYPE_FACILITY_REJ,          "Facility reject"},
  { MESSAGE_TYPE_LOOP_BACK_ACK,         "Loop back acknowledgement (national use)"},
  { MESSAGE_TYPE_PASS_ALONG,            "Pass-along (national use)"},
  { MESSAGE_TYPE_CIRC_GRP_RST_ACK,      "Circuit group reset acknowledgement"},
  { MESSAGE_TYPE_CIRC_GRP_QRY,          "Circuit group query (national use)"},
  { MESSAGE_TYPE_CIRC_GRP_QRY_RSP,      "Circuit group query response (national use)"},
  { MESSAGE_TYPE_CALL_PROGRSS,          "Call progress"},
  { MESSAGE_TYPE_USER2USER_INFO,        "User-to-user information"},
  { MESSAGE_TYPE_UNEQUIPPED_CIC,        "Unequipped CIC (national use)"},
  { MESSAGE_TYPE_CONFUSION,             "Confusion"},
  { MESSAGE_TYPE_OVERLOAD,              "Overload (national use)"},
  { MESSAGE_TYPE_CHARGE_INFO,           "Charge information (national use)"},
  { MESSAGE_TYPE_NETW_RESRC_MGMT,       "Network resource management"},
  { MESSAGE_TYPE_FACILITY,              "Facility"},
  { MESSAGE_TYPE_USER_PART_TEST,        "User part test"},
  { MESSAGE_TYPE_USER_PART_AVAIL,       "User part available"},
  { MESSAGE_TYPE_IDENT_REQ,             "Identification request"},
  { MESSAGE_TYPE_IDENT_RSP,             "Identification response"},
  { MESSAGE_TYPE_SEGMENTATION,          "Segmentation"},
  { MESSAGE_TYPE_LOOP_PREVENTION,       "Loop prevention"},
  { 0,                                  NULL}};

/* Same as above but in acronym form (for the Info column) */
static const value_string isup_message_type_value_acro[] = {
  { MESSAGE_TYPE_INITIAL_ADDR,          "IAM"},
  { MESSAGE_TYPE_SUBSEQ_ADDR,           "SAM"},
  { MESSAGE_TYPE_INFO_REQ,              "INR"},
  { MESSAGE_TYPE_INFO,                  "INF"},
  { MESSAGE_TYPE_CONTINUITY,            "COT"},
  { MESSAGE_TYPE_ADDR_CMPL,             "ACM"},
  { MESSAGE_TYPE_CONNECT,               "CON"},
  { MESSAGE_TYPE_FORW_TRANS,            "FOT"},
  { MESSAGE_TYPE_ANSWER,                "ANM"},
  { MESSAGE_TYPE_RELEASE,               "REL"},
  { MESSAGE_TYPE_SUSPEND,               "SUS"},
  { MESSAGE_TYPE_RESUME,                "RES"},
  { MESSAGE_TYPE_REL_CMPL,              "RLC"},
  { MESSAGE_TYPE_CONT_CHECK_REQ,        "CCR"},
  { MESSAGE_TYPE_RESET_CIRCUIT,         "RSC"},
  { MESSAGE_TYPE_BLOCKING,              "BLO"},
  { MESSAGE_TYPE_UNBLOCKING,            "UBL"},
  { MESSAGE_TYPE_BLOCK_ACK,             "BLA"},
  { MESSAGE_TYPE_UNBLOCK_ACK,           "UBLA"},
  { MESSAGE_TYPE_CIRC_GRP_RST,          "GRS"},
  { MESSAGE_TYPE_CIRC_GRP_BLCK,         "CGB"},
  { MESSAGE_TYPE_CIRC_GRP_UNBL,         "CGU"},
  { MESSAGE_TYPE_CIRC_GRP_BL_ACK,       "CGBA"},
  { MESSAGE_TYPE_CIRC_GRP_UNBL_ACK,     "CGUA"},
  { MESSAGE_TYPE_FACILITY_REQ,          "FAR"},
  { MESSAGE_TYPE_FACILITY_ACC,          "FAA"},
  { MESSAGE_TYPE_FACILITY_REJ,          "FRJ"},
  { MESSAGE_TYPE_LOOP_BACK_ACK,         "LPA"},
  { MESSAGE_TYPE_PASS_ALONG,            "PAM"},
  { MESSAGE_TYPE_CIRC_GRP_RST_ACK,      "GRA"},
  { MESSAGE_TYPE_CIRC_GRP_QRY,          "CQM"},
  { MESSAGE_TYPE_CIRC_GRP_QRY_RSP,      "CQR"},
  { MESSAGE_TYPE_CALL_PROGRSS,          "CPG"},
  { MESSAGE_TYPE_USER2USER_INFO,        "UUI"},
  { MESSAGE_TYPE_UNEQUIPPED_CIC,        "UCIC"},
  { MESSAGE_TYPE_CONFUSION,             "CFN"},
  { MESSAGE_TYPE_OVERLOAD,              "OLM"},
  { MESSAGE_TYPE_CHARGE_INFO,           "CRG"},
  { MESSAGE_TYPE_NETW_RESRC_MGMT,       "NRM"},
  { MESSAGE_TYPE_FACILITY,              "FAC"},
  { MESSAGE_TYPE_USER_PART_TEST,        "UPT"},
  { MESSAGE_TYPE_USER_PART_AVAIL,       "UPA"},
  { MESSAGE_TYPE_IDENT_REQ,             "IDR"},
  { MESSAGE_TYPE_IDENT_RSP,             "IDS"},
  { MESSAGE_TYPE_SEGMENTATION,          "SGM"},
  { MESSAGE_TYPE_LOOP_PREVENTION,       "LOP"},
  { 0,                                  NULL}};

/* Definition of Parameter Types */
#define PARAM_TYPE_END_OF_OPT_PARAMS      0
#define PARAM_TYPE_CALL_REF               1
#define PARAM_TYPE_TRANSM_MEDIUM_REQU     2
#define PARAM_TYPE_ACC_TRANSP             3
#define PARAM_TYPE_CALLED_PARTY_NR        4
#define PARAM_TYPE_SUBSQT_NR              5
#define PARAM_TYPE_NATURE_OF_CONN_IND     6
#define PARAM_TYPE_FORW_CALL_IND          7
#define PARAM_TYPE_OPT_FORW_CALL_IND      8
#define PARAM_TYPE_CALLING_PRTY_CATEG     9
#define PARAM_TYPE_CALLING_PARTY_NR      10
#define PARAM_TYPE_REDIRECTING_NR        11
#define PARAM_TYPE_REDIRECTION_NR        12
#define PARAM_TYPE_CONNECTION_REQ        13
#define PARAM_TYPE_INFO_REQ_IND          14
#define PARAM_TYPE_INFO_IND              15
#define PARAM_TYPE_CONTINUITY_IND        16
#define PARAM_TYPE_BACKW_CALL_IND        17
#define PARAM_TYPE_CAUSE_INDICATORS      18
#define PARAM_TYPE_REDIRECTION_INFO      19
#define PARAM_TYPE_CIRC_GRP_SV_MSG_TYPE  21
#define PARAM_TYPE_RANGE_AND_STATUS      22
#define PARAM_TYPE_FACILITY_IND          24
#define PARAM_TYPE_CLSD_USR_GRP_ILOCK_CD 26
#define PARAM_TYPE_USER_SERVICE_INFO     29
#define PARAM_TYPE_SIGNALLING_POINT_CODE 30
#define PARAM_TYPE_USER_TO_USER_INFO     32
#define PARAM_TYPE_CONNECTED_NR          33
#define PARAM_TYPE_SUSP_RESUME_IND       34
#define PARAM_TYPE_TRANSIT_NETW_SELECT   35
#define PARAM_TYPE_EVENT_INFO            36
#define PARAM_TYPE_CIRC_ASSIGN_MAP       37
#define PARAM_TYPE_CIRC_STATE_IND        38
#define PARAM_TYPE_AUTO_CONG_LEVEL       39
#define PARAM_TYPE_ORIG_CALLED_NR        40
#define PARAM_TYPE_OPT_BACKW_CALL_IND    41
#define PARAM_TYPE_USER_TO_USER_IND      42
#define PARAM_TYPE_ORIG_ISC_POINT_CODE   43
#define PARAM_TYPE_GENERIC_NOTIF_IND     44
#define PARAM_TYPE_CALL_HIST_INFO        45
#define PARAM_TYPE_ACC_DELIV_INFO        46
#define PARAM_TYPE_NETW_SPECIFIC_FACLTY  47
#define PARAM_TYPE_USER_SERVICE_INFO_PR  48
#define PARAM_TYPE_PROPAG_DELAY_COUNTER  49
#define PARAM_TYPE_REMOTE_OPERATIONS     50
#define PARAM_TYPE_SERVICE_ACTIVATION    51
#define PARAM_TYPE_USER_TELESERV_INFO    52
#define PARAM_TYPE_TRANSM_MEDIUM_USED    53
#define PARAM_TYPE_CALL_DIV_INFO         54
#define PARAM_TYPE_ECHO_CTRL_INFO        55
#define PARAM_TYPE_MSG_COMPAT_INFO       56
#define PARAM_TYPE_PARAM_COMPAT_INFO     57
#define PARAM_TYPE_MLPP_PRECEDENCE       58
#define PARAM_TYPE_MCID_REQ_IND          59
#define PARAM_TYPE_MCID_RSP_IND          60
#define PARAM_TYPE_HOP_COUNTER           61
#define PARAM_TYPE_TRANSM_MEDIUM_RQUR_PR 62
#define PARAM_TYPE_LOCATION_NR           63
#define PARAM_TYPE_REDIR_NR_RSTRCT       64
#define PARAM_TYPE_CALL_TRANS_REF        67
#define PARAM_TYPE_LOOP_PREV_IND         68
#define PARAM_TYPE_CALL_TRANS_NR         69
#define PARAM_TYPE_CCSS                  75
#define PARAM_TYPE_FORW_GVNS             76
#define PARAM_TYPE_BACKW_GVNS            77
#define PARAM_TYPE_REDIRECT_CAPAB        78
#define PARAM_TYPE_NETW_MGMT_CTRL        91
#define PARAM_TYPE_CORRELATION_ID       101
#define PARAM_TYPE_SCF_ID               102
#define PARAM_TYPE_CALL_DIV_TREAT_IND   110
#define PARAM_TYPE_CALLED_IN_NR         111
#define PARAM_TYPE_CALL_OFF_TREAT_IND   112
#define PARAM_TYPE_CHARGED_PARTY_IDENT  113
#define PARAM_TYPE_CONF_TREAT_IND       114
#define PARAM_TYPE_DISPLAY_INFO         115
#define PARAM_TYPE_UID_ACTION_IND       116
#define PARAM_TYPE_UID_CAPAB_IND        117
#define PARAM_TYPE_REDIRECT_COUNTER     119
#define PARAM_TYPE_COLLECT_CALL_REQ     121
#define PARAM_TYPE_GENERIC_NR           192
#define PARAM_TYPE_GENERIC_DIGITS       193

static const value_string isup_parameter_type_value[] = {
{ PARAM_TYPE_END_OF_OPT_PARAMS,        "End of optional parameters"},
  { PARAM_TYPE_CALL_REF,               "Call Reference (national use)"},
  { PARAM_TYPE_TRANSM_MEDIUM_REQU,     "Transmission medium requirement"},
  { PARAM_TYPE_ACC_TRANSP,             "Access transport"},
  { PARAM_TYPE_CALLED_PARTY_NR,        "Called party number"},
  { PARAM_TYPE_SUBSQT_NR,              "Subsequent number"},
  { PARAM_TYPE_NATURE_OF_CONN_IND,     "Nature of connection indicators"},
  { PARAM_TYPE_FORW_CALL_IND,          "Forward call indicators"},
  { PARAM_TYPE_OPT_FORW_CALL_IND,      "Optional forward call indicators"},
  { PARAM_TYPE_CALLING_PRTY_CATEG,     "Calling party's category"},
  { PARAM_TYPE_CALLING_PARTY_NR,       "Calling party number"},
  { PARAM_TYPE_REDIRECTING_NR,         "Redirecting number"},
  { PARAM_TYPE_REDIRECTION_NR,         "Redirection number"},
  { PARAM_TYPE_CONNECTION_REQ,         "Connection request"},
  { PARAM_TYPE_INFO_REQ_IND,           "Information request indicators (national use)"},
  { PARAM_TYPE_INFO_IND,               "Information indicators (national use)"},
  { PARAM_TYPE_CONTINUITY_IND,         "Continuity request"},
  { PARAM_TYPE_BACKW_CALL_IND,         "Backward call indicators"},
  { PARAM_TYPE_CAUSE_INDICATORS,       "Cause indicators"},
  { PARAM_TYPE_REDIRECTION_INFO,       "Redirection information"},
  { PARAM_TYPE_CIRC_GRP_SV_MSG_TYPE,   "Circuit group supervision message type"},
  { PARAM_TYPE_RANGE_AND_STATUS,       "Range and Status"},
  { PARAM_TYPE_FACILITY_IND,           "Facility indicator"},
  { PARAM_TYPE_CLSD_USR_GRP_ILOCK_CD,  "Closed user group interlock code"},
  { PARAM_TYPE_USER_SERVICE_INFO,      "User service information"},
  { PARAM_TYPE_SIGNALLING_POINT_CODE,  "Signalling point code (national use)"},
  { PARAM_TYPE_USER_TO_USER_INFO,      "User-to-user information"},
  { PARAM_TYPE_CONNECTED_NR,           "Connected number"},
  { PARAM_TYPE_SUSP_RESUME_IND,        "Suspend/Resume indicators"},
  { PARAM_TYPE_TRANSIT_NETW_SELECT,    "Transit network selection (national use)"},
  { PARAM_TYPE_EVENT_INFO,             "Event information"},
  { PARAM_TYPE_CIRC_ASSIGN_MAP,        "Circuit assignment map"},
  { PARAM_TYPE_CIRC_STATE_IND,         "Circuit state indicator (national use)"},
  { PARAM_TYPE_AUTO_CONG_LEVEL,        "Automatic congestion level"},
  { PARAM_TYPE_ORIG_CALLED_NR,         "Original called number"},
  { PARAM_TYPE_OPT_BACKW_CALL_IND,     "Backward call indicators"},
  { PARAM_TYPE_USER_TO_USER_IND,       "User-to-user indicators"},
  { PARAM_TYPE_ORIG_ISC_POINT_CODE,    "Origination ISC point code"},
  { PARAM_TYPE_GENERIC_NOTIF_IND,      "Generic notification indicator"},
  { PARAM_TYPE_CALL_HIST_INFO,         "Call history information"},
  { PARAM_TYPE_ACC_DELIV_INFO,         "Access delivery information"},
  { PARAM_TYPE_NETW_SPECIFIC_FACLTY,   "Network specific facility (national use)"},
  { PARAM_TYPE_USER_SERVICE_INFO_PR,   "User service information prime"},
  { PARAM_TYPE_PROPAG_DELAY_COUNTER,   "Propagation delay counter"},
  { PARAM_TYPE_REMOTE_OPERATIONS,      "Remote operations (national use)"},
  { PARAM_TYPE_SERVICE_ACTIVATION,     "Service activation"},
  { PARAM_TYPE_USER_TELESERV_INFO,     "User teleservice information"},
  { PARAM_TYPE_TRANSM_MEDIUM_USED,     "Transmission medium used"},
  { PARAM_TYPE_CALL_DIV_INFO,          "Call diversion information"},
  { PARAM_TYPE_ECHO_CTRL_INFO,         "Echo control information"},
  { PARAM_TYPE_MSG_COMPAT_INFO,        "Message compatibility information"},
  { PARAM_TYPE_PARAM_COMPAT_INFO,      "Parameter compatibility information"},
  { PARAM_TYPE_MLPP_PRECEDENCE,        "MLPP precedence"},
  { PARAM_TYPE_MCID_REQ_IND,           "MCID request indicators"},
  { PARAM_TYPE_MCID_RSP_IND,           "MCID response indicators"},
  { PARAM_TYPE_HOP_COUNTER,            "Hop counter"},
  { PARAM_TYPE_TRANSM_MEDIUM_RQUR_PR,  "Transmission medium requirement prime"},
  { PARAM_TYPE_LOCATION_NR,            "Location number"},
  { PARAM_TYPE_REDIR_NR_RSTRCT,        "Redirection number restriction"},
  { PARAM_TYPE_CALL_TRANS_REF,         "Call transfer reference"},
  { PARAM_TYPE_LOOP_PREV_IND,          "Loop prevention indicators"},
  { PARAM_TYPE_CALL_TRANS_NR,          "Call transfer number"},
  { PARAM_TYPE_CCSS,                   "CCSS"},
  { PARAM_TYPE_FORW_GVNS,              "Forward GVNS"},
  { PARAM_TYPE_BACKW_GVNS,             "Backward GVNS"},
  { PARAM_TYPE_REDIRECT_CAPAB,         "Redirect capability (reserved for national use)"},
  { PARAM_TYPE_NETW_MGMT_CTRL,         "Network management controls"},
  { PARAM_TYPE_CORRELATION_ID,         "Correlation id"},
  { PARAM_TYPE_SCF_ID,                 "SCF id"},
  { PARAM_TYPE_CALL_DIV_TREAT_IND,     "Call diversion treatment indicators"},
  { PARAM_TYPE_CALLED_IN_NR,           "Called IN number"},
  { PARAM_TYPE_CALL_OFF_TREAT_IND,     "Call offering treatment indicators"},
  { PARAM_TYPE_CHARGED_PARTY_IDENT,    "Charged party identification (national use)"},
  { PARAM_TYPE_CONF_TREAT_IND,         "Conference treatment indicators"},
  { PARAM_TYPE_DISPLAY_INFO,           "Display information"},
  { PARAM_TYPE_UID_ACTION_IND,         "UID action indicators"},
  { PARAM_TYPE_UID_CAPAB_IND,          "UID capability indicators"},
  { PARAM_TYPE_REDIRECT_COUNTER,       "Redirect counter (reserved for national use)"},
  { PARAM_TYPE_COLLECT_CALL_REQ,       "Collect call request"},
  { PARAM_TYPE_GENERIC_NR,             "Generic number"},
  { PARAM_TYPE_GENERIC_DIGITS,         "Generic digits (national use)"},
  { 0,                                 NULL}};


#define CIC_LENGTH                             2
#define MESSAGE_TYPE_LENGTH                    1
#define COMMON_HEADER_LENGTH                   (CIC_LENGTH + MESSAGE_TYPE_LENGTH)

#define MAXLENGTH                            0xFF /* since length field is 8 Bit long - used in number dissectors;
						     max. number of address digits is 15 digits, but MAXLENGTH used
						     to avoid runtime errors */

#define PARAMETER_TYPE_LENGTH                  1
#define PARAMETER_POINTER_LENGTH               1
#define PARAMETER_LENGTH_IND_LENGTH            1

/* All following parameter length definitions are WITHOUT the parameter type byte and length indicator for optional parameters*/
#define PARAMETER_NAME_LENGTH                  1
#define PARAMETER_LENGTH_IND_LENGTH            1
#define ACCESS_DELIVERY_INFO_LENGTH            1
#define AUTO_CONGEST_LEVEL_LENGTH              1
#define BACKWARD_CALL_IND_LENGTH               2
#define BACKWARD_GVNS_LENGTH                   1
#define CALL_DIV_INFO_LENGTH                   1
#define CALL_DIV_TREATMENT_IND_LENGTH          1
#define CALL_HISTORY_INFO_LENGTH               2
#define CALL_OFFERING_TREATMENT_IND_LENGTH     1
#define CALL_REFERENCE_LENGTH                  5
#define CALL_TRANSFER_REF_LENGTH               1
#define CALLING_PRTYS_CATEGORY_LENGTH          1
#define CCSS_LENGTH                            1
#define CIRCUIT_ASSIGNMENT_MAP_LENGTH          5
#define CIRC_GRP_SV_MSG_TYPE_LENGTH            1
#define CLOSED_USR_GRP_INTERLOCK_CODE_LENGTH   4
#define COLLECT_CALL_REQUEST_LENGTH            1
#define CONFERENCE_TREATMENT_IND_LENGTH        1
#define CONNECTION_REQUEST_LENGTH              7
#define CONTINUITY_IND_LENGTH                  1
#define ECHO_CONTROL_INFO_LENGTH               1
#define END_OF_OPT_PART_LENGTH                 1
#define EVENT_INFO_LENGTH                      1
#define FACILITY_IND_LENGTH                    1
#define FORWARD_CALL_IND_LENGTH                2
#define GENERIC_NOTIFICATION_IND_LENGTH        1
#define HOP_COUNTER_LENGTH                     1
#define INFO_IND_LENGTH                        2
#define INFO_REQUEST_IND_LENGTH                2
#define LOOP_PREVENTION_IND_LENGTH             1
#define MCID_REQUEST_IND_LENGTH                1
#define MCID_RESPONSE_IND_LENGTH               1
#define MLPP_PRECEDENCE_LENGTH                 1
#define NATURE_OF_CONNECTION_IND_LENGTH        1
#define NETWORK_MANAGEMENT_CONTROLS_LENGTH     1
#define OPTIONAL_BACKWARD_CALL_IND_LENGTH      1
#define OPTIONAL_FORWARD_CALL_IND_LENGTH       1
#define ORIGINAL_ISC_POINT_CODE_LENGTH         2
#define PROPAGATION_DELAY_COUNT_LENGTH         2
#define REDIRECTION_NUMBER_LENGTH              2
#define REDIRECTION_INFO_LENGTH                2
#define REDIRECTION_NUMBER_RESTRICTION_LENGTH  1
#define SIGNALLING_POINT_CODE_LENGTH           2
#define SUSPEND_RESUME_IND_LENGTH              1
#define TRANSMISSION_MEDIUM_REQUIREMENT_LENGTH 1
#define TRANSMISSION_MEDIUM_RQMT_PRIME_LENGTH  1
#define TRANSMISSION_MEDIUM_USED_LENGTH        1
#define UID_ACTION_IND_LENGTH                  1
#define UID_CAPABILITY_IND_LENGTH              1
#define USER_TELESERVICE_INFO_LENGTH           3
#define USER_TO_USER_IND_LENGTH                1
#define RANGE_LENGTH                           1

#define CALL_ID_LENGTH        3 /* for parameter Call Reference */
#define SPC_LENGTH            2 /* for parameter Call Reference, Connection request */
#define LOCAL_REF_LENGTH      3 /* for parameter Connection request */
#define PROTOCOL_CLASS_LENGTH 1 /* for parameter Connection request */
#define CREDIT_LENGTH         1 /* for parameter Connection request */

#define CIC_OFFSET            0

#define NO_SATELLITE_CIRCUIT_IN_CONNECTION   0
#define ONE_SATELLITE_CIRCUIT_IN_CONNECTION  1
#define TWO_SATELLITE_CIRCUIT_IN_CONNECTION  2
static const value_string isup_satellite_ind_value[] = {
  { NO_SATELLITE_CIRCUIT_IN_CONNECTION,          "No Satellite circuit in connection"},
  { ONE_SATELLITE_CIRCUIT_IN_CONNECTION,         "One Satellite circuit in connection"},
  { TWO_SATELLITE_CIRCUIT_IN_CONNECTION,         "Two Satellite circuits in connection"},
  { 0,                                 NULL}};

#define CONTINUITY_CHECK_NOT_REQUIRED           0
#define CONTINUITY_CHECK_REQUIRED               1
#define CONTINUITY_CHECK_ON_A_PREVIOUS_CIRCUIT  2
static const value_string isup_continuity_check_ind_value[] = {
  { CONTINUITY_CHECK_NOT_REQUIRED,               "Continuity check not required"},
  { CONTINUITY_CHECK_REQUIRED,                   "Continuity check required on this circuit"},
  { CONTINUITY_CHECK_ON_A_PREVIOUS_CIRCUIT ,     "Cont. check performed on a prev. circuit"},
  { 0,                                 NULL}};

static const true_false_string isup_echo_control_device_ind_value = {
  "Echo control device included",
  "Echo control device not included"
};

static const true_false_string isup_natnl_inatnl_call_ind_value = {
  "International call",
  "National call"
};

#define NO_END_TO_END_METHOD_AVAILABLE       0
#define PASS_ALONG_METHOD_AVAILABLE          1
#define SCCP_METHOD_AVAILABLE                2
#define PASS_ALONG_AND_SCCP_METHOD_AVAILABLE 3
static const value_string isup_end_to_end_method_ind_value[] = {
  { NO_END_TO_END_METHOD_AVAILABLE,          "No End-to-end method available(only link-by-link)"},
  { PASS_ALONG_METHOD_AVAILABLE,             "Pass-along method available (national use)"},
  { SCCP_METHOD_AVAILABLE,                   "SCCP method available"},
  { PASS_ALONG_AND_SCCP_METHOD_AVAILABLE,    "pass-along and SCCP method available(nat. use)"},
  { 0,                                 NULL}};

static const true_false_string isup_interworking_ind_value = {
  "interworking encountered",
  "no interworking encountered (SS7 all the way)"
};

static const true_false_string isup_end_to_end_info_ind_value = {
  "end-to-end information available",
  "no end-to-end information available"
};

static const true_false_string isup_ISDN_user_part_ind_value = {
  "ISDN user part used all the way",
  "ISDN user part not used all the way"
};

#define ISUP_PREFERED_ALL_THE_WAY               0
#define ISUP_NOT_REQUIRED_ALL_THE_WAY           1
#define ISUP_REQUIRED_ALL_WAY                   2
#define ISUP_ISDN_USER_PART_IND_SPARE           3
static const value_string isup_preferences_ind_value[] = {
  { ISUP_PREFERED_ALL_THE_WAY,                   "ISDN user part prefered all the way"},
  { ISUP_NOT_REQUIRED_ALL_THE_WAY,               "ISDN user part not required all the way"},
  { ISUP_REQUIRED_ALL_WAY,                       "ISDN user part required all the way"},
  { ISUP_ISDN_USER_PART_IND_SPARE,               "spare"},
  { 0,                                 NULL}};

static const true_false_string isup_ISDN_originating_access_ind_value = {
  "originating access ISDN",
  "originating access non-ISDN"
};

#define NO_INDICATION                                0
#define CONNECTIONLESS_METHOD_AVAILABLE              1
#define CONNECITON_ORIENTED_METHOD_AVAILABLE         2
#define CONNECTIONLESS_AND_ORIENTED_METHOD_AVAILABLE 3
static const value_string isup_SCCP_method_ind_value[] = {
  { NO_INDICATION,                                  "No indication"},
  { CONNECTIONLESS_METHOD_AVAILABLE,                "Connectionless method available (national use)"},
  { CONNECITON_ORIENTED_METHOD_AVAILABLE,           "Connection oriented method available"},
  { CONNECTIONLESS_AND_ORIENTED_METHOD_AVAILABLE,   "Connectionless and -oriented method avai.(nat. use)"},
  { 0,                                 NULL}};

#define UNKNOWN_AT_THIS_TIME                 0
#define OPERATOR_FRENCH                      1
#define OPERATOR_ENGLISH                     2
#define OPERATOR_GERMAN                      3
#define OPERATOR_RUSSIAN                     4
#define OPERATOR_SPANISH                     5
#define ORDINARY_CALLING_SUBSCRIBER         10
#define CALLING_SUBSCRIBER_WITH_PRIORITY    11
#define DATA_CALL                           12
#define TEST_CALL                           13
#define PAYPHONE                            15
static const value_string isup_calling_partys_category_value[] = {
  { UNKNOWN_AT_THIS_TIME,               "Category unknown at this time (national use)"},
  { OPERATOR_FRENCH,                    "operator, language French"},
  { OPERATOR_ENGLISH,                   "operator, language English"},
  { OPERATOR_GERMAN,                    "operator, language German"},
  { OPERATOR_RUSSIAN,                   "operator, language Russian"},
  { OPERATOR_SPANISH,                   "operator, language Spanish"},
  { ORDINARY_CALLING_SUBSCRIBER,        "ordinary calling subscriber"},
  { CALLING_SUBSCRIBER_WITH_PRIORITY,   "calling subscriber with priority"},
  { DATA_CALL,                          "data call (voice band data)"},
  { TEST_CALL,                          "test call"},
  { PAYPHONE,                           "payphone"},
  { 0,                                 NULL}};

#define MEDIUM_SPEECH                        0
#define MEDIUM_64KBS                         2
#define MEDIUM_3_1_KHZ_AUDIO                 3
#define MEDIUM_RESERVED_SERVICE2_1           4
#define MEDIUM_RESERVED_SERVICE1_2           5
#define MEDIUM_64KBS_PREFERED                6
#define MEDIUM_2_64KBS                       7
#define MEDIUM_384KBS                        8
#define MEDIUM_1536KBS                       9
#define MEDIUM_1920KBS                      10
#define MEDIUM_3_64KBS                      16
#define MEDIUM_4_64KBS                      17
#define MEDIUM_5_64KBS                      18
#define MEDIUM_7_64KBS                      20
#define MEDIUM_8_64KBS                      21
#define MEDIUM_9_64KBS                      22
#define MEDIUM_10_64KBS                     23
#define MEDIUM_11_64KBS                     24
#define MEDIUM_12_64KBS                     25
#define MEDIUM_13_64KBS                     26
#define MEDIUM_14_64KBS                     27
#define MEDIUM_15_64KBS                     28
#define MEDIUM_16_64KBS                     29
#define MEDIUM_17_64KBS                     30
#define MEDIUM_18_64KBS                     31
#define MEDIUM_19_64KBS                     32
#define MEDIUM_20_64KBS                     33
#define MEDIUM_21_64KBS                     34
#define MEDIUM_22_64KBS                     35
#define MEDIUM_23_64KBS                     36
#define MEDIUM_25_64KBS                     38
#define MEDIUM_26_64KBS                     39
#define MEDIUM_27_64KBS                     40
#define MEDIUM_28_64KBS                     41
#define MEDIUM_29_64KBS                     42
static const value_string isup_transmission_medium_requirement_value[] = {
  { MEDIUM_SPEECH,                       "speech"},
  { MEDIUM_64KBS,                        "64 kbit/s unrestricted"},
  { MEDIUM_3_1_KHZ_AUDIO,                "3.1 kHz audio"},
  { MEDIUM_RESERVED_SERVICE2_1,          "reserved for alternate speech (service 2)/64 kbit/s unrestricted (service 1)"},
  { MEDIUM_RESERVED_SERVICE1_2,          "reserved for alternate 64 kbit/s unrestricted (service 1)/speech (service 2)"},
  { MEDIUM_64KBS_PREFERED,               "64 kbit/s prefered"},
  { MEDIUM_2_64KBS,                      "2x64 kbit/s unrestricted"},
  { MEDIUM_384KBS,                       "384 kbit/s unrestricted"},
  { MEDIUM_1536KBS,                      "1536 kbit/s unrestricted"},
  { MEDIUM_1920KBS,                      "1920 kbit/s unrestricted"},
  { MEDIUM_3_64KBS,                      "3x64 kbit/s unrestricted"},
  { MEDIUM_4_64KBS,                      "4x64 kbit/s unrestricted"},
  { MEDIUM_5_64KBS,                      "5x64 kbit/s unrestricted"},
  { MEDIUM_7_64KBS,                      "7x64 kbit/s unrestricted"},
  { MEDIUM_8_64KBS,                      "8x64 kbit/s unrestricted"},
  { MEDIUM_9_64KBS,                      "9x64 kbit/s unrestricted"},
  { MEDIUM_10_64KBS,                     "10x64 kbit/s unrestricted"},
  { MEDIUM_11_64KBS,                     "11x64 kbit/s unrestricted"},
  { MEDIUM_12_64KBS,                     "12x64 kbit/s unrestricted"},
  { MEDIUM_13_64KBS,                     "13x64 kbit/s unrestricted"},
  { MEDIUM_14_64KBS,                     "14x64 kbit/s unrestricted"},
  { MEDIUM_15_64KBS,                     "15x64 kbit/s unrestricted"},
  { MEDIUM_16_64KBS,                     "16x64 kbit/s unrestricted"},
  { MEDIUM_17_64KBS,                     "17x64 kbit/s unrestricted"},
  { MEDIUM_18_64KBS,                     "18x64 kbit/s unrestricted"},
  { MEDIUM_19_64KBS,                     "19x64 kbit/s unrestricted"},
  { MEDIUM_20_64KBS,                     "20x64 kbit/s unrestricted"},
  { MEDIUM_21_64KBS,                     "21x64 kbit/s unrestricted"},
  { MEDIUM_22_64KBS,                     "22x64 kbit/s unrestricted"},
  { MEDIUM_23_64KBS,                     "23x64 kbit/s unrestricted"},
  { MEDIUM_25_64KBS,                     "25x64 kbit/s unrestricted"},
  { MEDIUM_26_64KBS,                     "26x64 kbit/s unrestricted"},
  { MEDIUM_27_64KBS,                     "27x64 kbit/s unrestricted"},
  { MEDIUM_28_64KBS,                     "28x64 kbit/s unrestricted"},
  { MEDIUM_29_64KBS,                     "29x64 kbit/s unrestricted"},
  { 0,                                 NULL}};
static const value_string isup_transmission_medium_requirement_prime_value[] = {
  { MEDIUM_SPEECH,                       "speech"},
  { MEDIUM_64KBS,                        "reserved for 64 kbit/s unrestricted"},
  { MEDIUM_3_1_KHZ_AUDIO,                "3.1 kHz audio"},
  { MEDIUM_RESERVED_SERVICE2_1,          "reserved for alternate speech (service 2)/64 kbit/s unrestricted (service 1)"},
  { MEDIUM_RESERVED_SERVICE1_2,          "reserved for alternate 64 kbit/s unrestricted (service 1)/speech (service 2)"},
  { MEDIUM_64KBS_PREFERED,               "reserved for 64 kbit/s prefered"},
  { MEDIUM_2_64KBS,                      "reserved for 2x64 kbit/s unrestricted"},
  { MEDIUM_384KBS,                       "reserved for 384 kbit/s unrestricted"},
  { MEDIUM_1536KBS,                      "reserved for 1536 kbit/s unrestricted"},
  { MEDIUM_1920KBS,                      "reserved for 1920 kbit/s unrestricted"},
  { 0,                                 NULL}};


/* Definitions for Called and Calling Party number */
#define ISUP_ODD_EVEN_MASK                       0x80
#define ISUP_NATURE_OF_ADDRESS_IND_MASK          0x7F
#define ISUP_INN_MASK                            0x80
#define ISUP_NI_MASK                             0x80
#define ISUP_NUMBERING_PLAN_IND_MASK             0x70
#define ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK 0x0C
#define ISUP_SCREENING_IND_MASK                  0x03
#define ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK       0x0F
#define ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK      0xF0

static const true_false_string isup_odd_even_ind_value = {
  "odd number of address signals",
  "even number of address signals"
};

#define ISUP_CALLED_PARTY_NATURE_SUBSCRIBER_NR  1
#define ISUP_CALLED_PARTY_NATURE_UNKNOWN        2
#define ISUP_CALLED_PARTY_NATURE_NATIONAL_NR    3
#define ISUP_CALLED_PARTY_NATURE_INTERNATNL_NR  4
#define ISUP_CALLED_PARTY_NATURE_NETW_SPEC_NR   5
static const value_string isup_called_party_nature_of_address_ind_value[] = {
  { ISUP_CALLED_PARTY_NATURE_SUBSCRIBER_NR,     "subscriber number (national use)"},
  { ISUP_CALLED_PARTY_NATURE_UNKNOWN,           "unknown (national use)"},
  { ISUP_CALLED_PARTY_NATURE_NATIONAL_NR,       "national (significant) number"},
  { ISUP_CALLED_PARTY_NATURE_INTERNATNL_NR,     "international number"},
  { ISUP_CALLED_PARTY_NATURE_NETW_SPEC_NR,      "network-specific number (national use)"},
  { 0,                                 NULL}};

static const value_string isup_calling_party_nature_of_address_ind_value[] = {
  { ISUP_CALLED_PARTY_NATURE_SUBSCRIBER_NR,     "subscriber number (national use)"},
  { ISUP_CALLED_PARTY_NATURE_UNKNOWN,           "unknown (national use)"},
  { ISUP_CALLED_PARTY_NATURE_NATIONAL_NR,       "national (significant) number"},
  { ISUP_CALLED_PARTY_NATURE_INTERNATNL_NR,     "international number"},
  { 0,                                 NULL}};

static const true_false_string isup_INN_ind_value = {
  "routing to internal network number not allowed",
  "routing to internal network number allowed "
};
static const true_false_string isup_NI_ind_value = {
  "incomplete",
  "complete"
};

#define ISDN_NUMBERING_PLAN                     1
#define DATA_NUMBERING_PLAN                     3
#define TELEX_NUMBERING_PLAN                    4
static const value_string isup_numbering_plan_ind_value[] = {
  { ISDN_NUMBERING_PLAN,     "ISDN (Telephony) numbering plan"},
  { DATA_NUMBERING_PLAN,     "Data numbering plan (national use)"},
  { TELEX_NUMBERING_PLAN,    "Telex numbering plan (national use)"},
  { 5,                       "Reserved for national use"},
  { 6,                       "Reserved for national use"},
  { 0,                                 NULL}};

#define ADDRESS_PRESETATION_ALLOWED      0
#define ADDRESS_PRESETATION_RESTRICTED   1
#define ADDRESS_NOT_AVAILABLE            2
static const value_string isup_address_presentation_restricted_ind_value[] = {
  { ADDRESS_PRESETATION_ALLOWED,     "presentation allowed"},
  { ADDRESS_PRESETATION_RESTRICTED,  "presentation restricted"},
  { ADDRESS_NOT_AVAILABLE,           "address not availabe (national use)"},
  { 3,                                 "spare"},
  { 0,                                 NULL}};

static const value_string isup_screening_ind_value[] = {
  { 0,     "reserved"},
  { 1,     "user provided, verified and passed"},
  { 2,     "reserved"},
  { 3,     "network provided"},
  { 0,     NULL}};

static const value_string isup_screening_ind_enhanced_value[] = {
  { 0,     "user provided, not verified"},
  { 1,     "user provided, verified and passed"},
  { 2,     "user provided, verified and failed"},
  { 3,     "network provided"},
  { 0,     NULL}};

static const value_string isup_called_party_address_digit_value[] = {
  { 0,  "0"},
  { 1,  "1"},
  { 2,  "2"},
  { 3,  "3"},
  { 4,  "4"},
  { 5,  "5"},
  { 6,  "6"},
  { 7,  "7"},
  { 8,  "8"},
  { 9,  "9"},
  { 10, "spare"},
  { 11, "code 11 "},
  { 12, "code 12"},
  { 15, "Stop sending"},
  { 0,  NULL}};

static const value_string isup_calling_party_address_digit_value[] = {
  { 0,  "0"},
  { 1,  "1"},
  { 2,  "2"},
  { 3,  "3"},
  { 4,  "4"},
  { 5,  "5"},
  { 6,  "6"},
  { 7,  "7"},
  { 8,  "8"},
  { 9,  "9"},
  { 10, "spare"},
  { 11, "code 11 "},
  { 12, "code 12"},
  { 15, "spare"},
  { 0,  NULL}};

/*End of Called/Calling party address definitions */


static const true_false_string isup_calling_party_address_request_ind_value = {
  "calling party address requested",
  "calling party address not requested"
};
static const true_false_string isup_holding_ind_value = {
  "holding requested",
  "holding not requested"
};
static const true_false_string isup_calling_partys_category_request_ind_value = {
  "Calling Party's category requested",
  "Calling Party's category not requested",
};
static const true_false_string isup_charge_information_request_ind_value = {
  "Charge Information requested",
  "Charge Information not requested"
};
static const true_false_string isup_malicious_call_identification_request_ind_value = {
  "Malicious call identification requested",
  "Malicious call identification not requested"
};

#define CALLING_PARTY_ADDRESS_NOT_INCLUDED             0
#define CALLING_PARTY_ADDRESS_NOT_AVAILABLE            1
#define CALLING_PARTY_ADDRESS_INCLUDED                 3
static const value_string isup_calling_party_address_response_ind_value[] = {
  { CALLING_PARTY_ADDRESS_NOT_INCLUDED, "Calling party address not included"},
  { CALLING_PARTY_ADDRESS_NOT_AVAILABLE,"Calling party address not available"},
  { 4,                                  "spare"},
  { CALLING_PARTY_ADDRESS_INCLUDED,     "Calling party address included"},
  { 0,                                 NULL}};

static const true_false_string isup_hold_provided_ind_value = {
  "hold provided",
  "hold not provided"
};
static const true_false_string isup_calling_partys_category_response_ind_value = {
  "Calling Party's category included",
  "Calling Party's category not included",
};
static const true_false_string isup_charge_information_response_ind_value = {
  "Charge Information included",
  "Charge Information not included"
};
static const true_false_string isup_solicited_information_ind_value = {
  "unsolicited",
  "solicited"
};

static const true_false_string isup_continuity_ind_value = {
  "Continuity check successful",
  "Continuity ckec failed"
};

#define CHARGE_NO_IND       0
#define CHARGE_NO_CHARGE    1
#define CHARGE_CHARGE       2
static const value_string isup_charge_ind_value[] = {
  { CHARGE_NO_IND,    "No indication"},
  { CHARGE_NO_CHARGE, "No charge"},
  { CHARGE_CHARGE,    "Charge"},
  { 3,                "spare"},
  { 0,                NULL}};

#define CALLED_PARTYS_STATUS_NO_IND            0
#define CALLED_PARTYS_STATUS_SUBSCR_FREE       1
#define CALLED_PARTYS_STATUS_CONNECT_WHEN_FREE 2
static const value_string isup_called_partys_status_ind_value[] = {
  { CALLED_PARTYS_STATUS_NO_IND,            "No indication"},
  { CALLED_PARTYS_STATUS_SUBSCR_FREE,       "Subscriber free"},
  { CALLED_PARTYS_STATUS_CONNECT_WHEN_FREE, "Connect when free (national use)"},
  { 3,                                      "spare"},
  { 0,                NULL}};

#define CALLED_PARTYS_CATEGORY_NO_IND            0
#define CALLED_PARTYS_CATEGORY_ORDINARY_SUBSCR   1
#define CALLED_PARTYS_CATEGORY_PAYPHONE          2
static const value_string isup_called_partys_category_ind_value[] = {
  { CALLED_PARTYS_CATEGORY_NO_IND,             "No indication"},
  { CALLED_PARTYS_CATEGORY_ORDINARY_SUBSCR,    "Ordinary subscriber"},
  { CALLED_PARTYS_CATEGORY_PAYPHONE,           "Payphone"},
  { 3,                                         "spare"},
  { 0,                NULL}};

static const true_false_string isup_ISDN_terminating_access_ind_value = {
  "terminating access ISDN",
  "terminating access non-ISDN"
};

static const true_false_string isup_suspend_resume_ind_value = {
  "network initiated",
  "ISDN subscriber initiated"
};
#define MAINTENANCE            0
#define HARDWARE_FAILURE       1
#define RES_FOR_NATNL_USE      2
static const value_string isup_cgs_message_type_value[] = {
  { MAINTENANCE,            "maintenance oriented"},
  { HARDWARE_FAILURE,       "hardware failure oriented"},
  { RES_FOR_NATNL_USE,      "reserved for national use (ISUP'84)"},
  { 3,                      "spare"},
  { 0,                NULL}};

#define USER_TO_USER_SERVICE      2
static const value_string isup_facility_ind_value[] = {
  { USER_TO_USER_SERVICE,     "user-to-user service"},
  { 0,                NULL}};

#define MTC_BLCK_STATE_TRANSIENT  0
#define MTC_BLCK_STATE_UNEQUIPPED 3
static const value_string isup_mtc_blocking_state_DC00_value[] = {
  {  MTC_BLCK_STATE_TRANSIENT,  "transient"},
  { 1,                          "spare"},
  { 2,                          "spare"},
  {  MTC_BLCK_STATE_UNEQUIPPED, "unequipped"},
  { 0,                NULL}};

#define MTC_BLCK_NO_BLOCKING      0
#define MTC_LOCALLY_BLOCKED       1
#define MTC_REMOTELY_BLOCKED      2
#define MTC_LOCAL_REMOTE_BLOCKED  3
static const value_string isup_mtc_blocking_state_DCnot00_value[] = {
  {  MTC_BLCK_NO_BLOCKING,     "no blocking (active)"},
  {  MTC_LOCALLY_BLOCKED,      "locally blocked"},
  {  MTC_REMOTELY_BLOCKED,     "remotely blocked"},
  {  MTC_LOCAL_REMOTE_BLOCKED, "locally and remotely blocked"},
  {  0,                NULL}};

#define CALL_PROC_INCOMING_BUSY 1
#define CALL_PROC_OUTGOING_BUSY 2
#define CALL_PROC_IDLE          3
static const value_string isup_call_processing_state_value[] = {
  {  CALL_PROC_INCOMING_BUSY,     "circuit incoming busy"},
  {  CALL_PROC_OUTGOING_BUSY,     "circuit outgoing busy"},
  {  CALL_PROC_IDLE,              "idle"},
  {  0,                NULL}};

#define HW_BLCK_NO_BLOCKING          0
#define HW_LOCALLY_BLOCKED           1
#define HW_REMOTELY_BLOCKED          2
#define HW_LOCAL_REMOTE_BLOCKED      3
static const value_string isup_HW_blocking_state_value[] = {
  {  HW_BLCK_NO_BLOCKING,     "no blocking (active)"},
  {  HW_LOCALLY_BLOCKED,      "locally blocked"},
  {  HW_REMOTELY_BLOCKED,     "remotely blocked"},
  {  HW_LOCAL_REMOTE_BLOCKED, "locally and remotely blocked"},
  {  0,                NULL}};

#define EVENT_ALERTING      1
#define EVENT_PROGRESS      2
#define EVENT_INBAND_INFO   3
#define EVENT_ON_BUSY       4
#define EVENT_ON_NO_REPLY   5
#define EVENT_UNCONDITIONAL 6
static const value_string isup_event_ind_value[] = {
  /* according 3.21/Q.763 */
  {  EVENT_ALERTING,     "ALERTING"},
  {  EVENT_PROGRESS,     "PROGRESS"},
  {  EVENT_INBAND_INFO,  "in-band info. or an appropriate pattern is now avai."},
  {  EVENT_ON_BUSY,      "call forwarded on busy (national use)"},
  {  EVENT_ON_NO_REPLY,  "call forwarded on no reply (national use)"},
  {  EVENT_UNCONDITIONAL,"call forwarded unconditional (national use)"},
  {  0,                NULL}};

static const true_false_string isup_event_presentation_restricted_ind_value = {
  /* according 3.21/Q.763 */
  "presentation restricted",
  "no indication"
};
#define CUG_NON_CUG_CALL                      0
#define CUG_CALL_OUTGOING_ACCESS_ALLOWED      2
#define CUG_CALL_OUTGOING_ACCESS_NOT_ALLOWED  3
static const value_string isup_CUG_call_ind_value[] = {
  /* according 3.38/Q.763 */
  {  CUG_NON_CUG_CALL,                     "non-CUG call"},
  {  1,                                    "spare"},
  {  CUG_CALL_OUTGOING_ACCESS_ALLOWED,     "closed user group call, outgoing access allowed"},
  {  CUG_CALL_OUTGOING_ACCESS_NOT_ALLOWED, "closed user group call, outgoing access not allowed"},
  {  0,                NULL}};


static const true_false_string isup_simple_segmentation_ind_value = {
  /* according 3.38/Q.763 */
  "additional information will be sent in a segmentation message",
  "no additional information will be sent"
};

static const true_false_string isup_connected_line_identity_request_ind_value = {
  /* according 3.38/Q.763 */
  "requested",
  "not requested"
};

static const value_string isup_redirecting_ind_value[] = {
  /* according 3.45/Q.763 */
  {  0,        "no redirection (national use)"},
  {  1,        "call rerouted (national use)"},
  {  2,        "call rerouted, all redirection information presentation restricted (national use)"},
  {  3,        "call diverted"},
  {  4,        "call diverted, all redirection information presentation restricted"},
  {  5,        "call rerouted, redirection number presentation restricted (national use)"},
  {  6,        "call diversion, redirection number presentation restricted (national use)"},
  {  7,        "spare"},
  {  0,         NULL}};

static const value_string isup_original_redirection_reason_value[] = {
  /* according 3.45/Q.763 */
  {  0,        "unknown/not available"},
  {  1,        "user busy (national use)"},
  {  2,        "no reply (national use)"},
  {  3,        "unconditional (national use)"},
  {  0,         NULL}};

static const value_string isup_redirection_reason_value[] = {
  /* according 3.45/Q.763 */
  {  0,        "unknown/not available"},
  {  1,        "user busy (national use)"},
  {  2,        "no reply (national use)"},
  {  3,        "unconditional (national use)"},
  {  4,        "deflection during alerting"},
  {  5,        "deflection immediate response"},
  {  6,        "mobile subscriber not reachable"},
  {  0,         NULL}};

static const value_string isup_type_of_network_identification_value[] = {
  /* according 3.53/Q.763 */
  {  0,        "CCITT/ITU-T-standardized identification"},
  {  2,        "national network identification"},
  {  0,         NULL}};

static const value_string isup_network_identification_plan_value[] = {
  /* according 3.53/Q.763 */
  {  0,        "if CCITT/ITU-T id - unknown"},
  {  3,        "if CCITT/ITU-T id - public data network id code (X.121)"},
  {  6,        "if CCITT/ITU-T id - public land Mobile Network id code (E.211)"},
  {  0,         NULL}};

static const value_string isup_map_type_value[] = {
  /* according 3.69/Q.763 */
  {  1,        "1544 kbit/s digital path map format (64 kbit/s base rate"},
  {  2,        "2048 kbit/s digital path map format (64 kbit/s base rate"},
  {  0,         NULL}};

static const value_string isup_auto_congestion_level_value[] = {
  /* according 3.4/Q.763 */
  {  1,        "Congestion level 1 exceeded"},
  {  2,        "Congestion level 2 exceeded"},
  {  0,         NULL}};

static const true_false_string isup_inband_information_ind_value = {
  /* according 3.37/Q.763 */
  "in-band info. or an appropirate pattern is now avai.",
  "no indication"
};
static const true_false_string isup_call_diversion_may_occur_ind_value = {
  /* according 3.37/Q.763 */
  "call diversion may occur",
  "no indication"
};
static const true_false_string isup_MLPP_user_ind_value = {
  /* according 3.37/Q.763 */
  "MLPP user",
  "no indication"
};

static const true_false_string isup_access_delivery_ind_value = {
  /* according 3.2/Q.763 */
  "No set-up message generated",
  "Set-up message generated"
};

static const value_string isup_loop_prevention_response_ind_value[] = {
  /* according 3.67/Q.763 */
  {  0,        "insufficient information"},
  {  1,        "no loop exists"},
  {  2,        "simultaneous transfer"},
  {  0,         NULL}};

static const true_false_string isup_temporary_alternative_routing_ind_value = {
  /* according 3.68/Q.763 */
  "TAR controlled call",
  "no indication"
};
static const true_false_string isup_extension_ind_value = {
  /* according 3.68/Q.763 */
  "last octet",
  "information continues through the next octet"
};

static const value_string isup_call_to_be_diverted_ind_value[] = {
  /* according 3.72/Q.763 */
  {  0,        "no indication"},
  {  1,        "call diversion allowed"},
  {  2,        "call diversion not allowed"},
  {  3,        "spare"},
  {  0,         NULL}};

static const value_string isup_call_to_be_offered_ind_value[] = {
  /* according 3.72/Q.763 */
  {  0,        "no indication"},
  {  1,        "call offering not allowed"},
  {  2,        "call offering allowed"},
  {  3,        "spare"},
  {  0,         NULL}};

static const value_string isup_conference_acceptance_ind_value[] = {
  /* according 3.76/Q.763 */
  {  0,        "no indication"},
  {  1,        "accept conference request"},
  {  2,        "reject conference request"},
  {  3,        "spare"},
  {  0,         NULL}};


/* Generalized bit masks for 8 and 16 bits fields */
#define A_8BIT_MASK  0x01
#define B_8BIT_MASK  0x02
#define C_8BIT_MASK  0x04
#define D_8BIT_MASK  0x08
#define E_8BIT_MASK  0x10
#define F_8BIT_MASK  0x20
#define G_8BIT_MASK  0x40
#define H_8BIT_MASK  0x80

#define BA_8BIT_MASK 0x03
#define CB_8BIT_MASK 0x06
#define DC_8BIT_MASK 0x0C
#define FE_8BIT_MASK 0x30
#define GFE_8BIT_MASK 0x70
#define DCBA_8BIT_MASK 0x0F
#define EDCBA_8BIT_MASK 0x1F
#define HGFE_8BIT_MASK 0xF0
#define GFEDCBA_8BIT_MASK 0x7F
#define FEDCBA_8BIT_MASK 0x3F

#define A_16BIT_MASK  0x0100
#define B_16BIT_MASK  0x0200
#define C_16BIT_MASK  0x0400
#define D_16BIT_MASK  0x0800
#define E_16BIT_MASK  0x1000
#define F_16BIT_MASK  0x2000
#define G_16BIT_MASK  0x4000
#define H_16BIT_MASK  0x8000
#define I_16BIT_MASK  0x0001
#define J_16BIT_MASK  0x0002
#define K_16BIT_MASK  0x0004
#define L_16BIT_MASK  0x0008
#define M_16BIT_MASK  0x0010
#define N_16BIT_MASK  0x0020
#define O_16BIT_MASK  0x0040
#define P_16BIT_MASK  0x0080

#define BA_16BIT_MASK 0x0300
#define CB_16BIT_MASK 0x0600
#define DC_16BIT_MASK 0x0C00
#define FE_16BIT_MASK 0x3000
#define HG_16BIT_MASK 0xC000
#define KJ_16BIT_MASK 0x0006
#define PO_16BIT_MASK 0x00C0

#define CBA_16BIT_MASK 0x0700
#define KJI_16BIT_MASK 0x0007
#define HGFE_16BIT_MASK 0xF000
#define PONM_16BIT_MASK 0x00F0



/* ------------------------------------------------------------------
  Mapping number to ASCII-character
 ------------------------------------------------------------------ */
char number_to_char(int number)
{
  if (number < 10)
    return ((char) number + ASCII_NUMBER_DELTA);
  else
    return ((char) number + ASCII_LETTER_DELTA);
}


void print_parameter_title(int offset, int len, int parameter_type)
{
	  static char str_parameter[128];
	  sprintf(&str_parameter[0], "%s   Length: %u Data: 0x%s", val_to_str(parameter_type,isup_parameter_type_value,"unknown"), len, tvb_bytes_to_str(offset, len ) );
	  proto_tree_add_text(1, "%s", &str_parameter[0]);
}




/* ------------------------------------------------------------------ */
/* Dissectors for all used parameter types                            */
/* ------------------------------------------------------------------ */
/* argument tvbuff_t contains only parameter-specific length          */
/* length indicator is already dissected in dissect_isup_message() or */
/* dissect_isup_optional_parameter()                                  */
/* ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
 Dissector Parameter nature of connection flags
 */
static void
dissect_isup_nature_of_connection_indicators_parameter(int offset, int len)
{
  unsigned char nature_of_connection_ind;
  nature_of_connection_ind = tvb_get_guint8(offset);
  proto_tree_add_text(2, "Satellite Indicator: %s", val_to_str( (nature_of_connection_ind & BA_8BIT_MASK), isup_satellite_ind_value, NULL));
  proto_tree_add_text(2, "Continuity Check Indicator: %s", val_to_str( (nature_of_connection_ind & DC_8BIT_MASK) >> 2, isup_continuity_check_ind_value, NULL));
  proto_tree_add_text(2, "Echo Control Device Indicator: %s", tf_to_str( (nature_of_connection_ind & E_8BIT_MASK), isup_echo_control_device_ind_value));

}
/* ------------------------------------------------------------------
 Dissector Parameter Forward Call Indicators
 */
static void
dissect_isup_forward_call_indicators_parameter(int offset, int len)
{
  int forward_call_ind;

  forward_call_ind = tvb_get_guint16(offset);
//  proto_tree_add_text_i(1, "Forward Call Indicators    Data: 0x%x", forward_call_ind);
 
  proto_tree_add_text(2, "National/international call indicator: %s", tf_to_str( (forward_call_ind & A_16BIT_MASK), isup_natnl_inatnl_call_ind_value));
  proto_tree_add_text(2, "End-to-end method ind.: %s", val_to_str( (forward_call_ind & CB_16BIT_MASK) >> 9, isup_end_to_end_method_ind_value, NULL));
  proto_tree_add_text(2, "Interworking indicator: %s", tf_to_str( (forward_call_ind & D_16BIT_MASK) >> 11, isup_interworking_ind_value));
  proto_tree_add_text(2, "End-to-end information ind.: %s", tf_to_str( (forward_call_ind & E_16BIT_MASK) >> 12, isup_end_to_end_info_ind_value));
  proto_tree_add_text(2, "ISDN user part indicator: %s", tf_to_str( (forward_call_ind & F_16BIT_MASK) >> 13, isup_ISDN_user_part_ind_value));
  proto_tree_add_text(2, "ISDN user part preference ind.: %s", val_to_str( (forward_call_ind & HG_16BIT_MASK) >> 14, isup_preferences_ind_value, NULL));
  proto_tree_add_text(2, "ISDN access indicator: %s", tf_to_str( (forward_call_ind & I_16BIT_MASK) , isup_ISDN_originating_access_ind_value));
  proto_tree_add_text(2, "SCCP method indicator: %s", val_to_str( (forward_call_ind & KJ_16BIT_MASK) >> 1, isup_SCCP_method_ind_value, NULL));
 
}

/* ------------------------------------------------------------------
 Dissector Parameter Calling Party's Category
 */
static void
dissect_isup_calling_partys_category_parameter(int offset, int len)
{
  unsigned char calling_partys_category;

  calling_partys_category = tvb_get_guint8(offset);
 
  proto_tree_add_text(2, "Calling Party's category: %s", val_to_str( calling_partys_category, isup_calling_partys_category_value, NULL));
}


/* ------------------------------------------------------------------
  Dissector Parameter Transmission medium requirement
 */
static void
dissect_isup_transmission_medium_requirement_parameter(int offset, int len)
{
  unsigned char transmission_medium_requirement;

  transmission_medium_requirement = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Transmission medium requirement: %s", val_to_str( transmission_medium_requirement, isup_transmission_medium_requirement_value, NULL));
}

/* ------------------------------------------------------------------
  Dissector Parameter Called party number
 */
static void
dissect_isup_called_party_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char called_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_called_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "INN indicator: %s", tf_to_str( (indicators2 & ISUP_INN_MASK) , isup_INN_ind_value));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    called_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  called_number[i++] = '\0';
  proto_tree_add_text(2, "Called Party Number: %s", called_number);
}

/* ------------------------------------------------------------------
  Dissector Parameter  Subsequent number
 */
static void
dissect_isup_subsequent_number_parameter(int offset, int len)
{
  unsigned char indicators1;
  unsigned char address_digit_pair=0;
  int i=0;
  char called_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  len -= 1;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    called_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if ( (indicators1 & 0x80) == 0) { /* Even Indicator set -> last even digit is valid & has be displayed */
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
  }
  called_number[i++] = '\0';

  proto_tree_add_text(2, "Subsequent Number: %s", called_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Information Request Indicators
 */
static void
dissect_isup_information_request_indicators_parameter(int offset, int len)
{
  int information_request_indicators;

  information_request_indicators = tvb_get_guint16(offset);

  proto_tree_add_text(2, "Calling party address request indicator: %s", tf_to_str( (information_request_indicators & A_16BIT_MASK) >> 8, isup_calling_party_address_request_ind_value));
  proto_tree_add_text(2, "Holding indicator: %s", tf_to_str( (information_request_indicators & B_16BIT_MASK) >> 9, isup_holding_ind_value));
  proto_tree_add_text(2, "Calling party's category request indicator: %s", tf_to_str( (information_request_indicators & D_16BIT_MASK) >> 11, isup_calling_partys_category_request_ind_value));
  proto_tree_add_text(2, "Charge information request indicator: %s", tf_to_str( (information_request_indicators & E_16BIT_MASK) >> 12, isup_charge_information_request_ind_value));
  proto_tree_add_text(2, "Malicious call identification request indicator (ISUP'88): %s", tf_to_str( (information_request_indicators & H_16BIT_MASK) >> 15, isup_malicious_call_identification_request_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter Information Indicators
 */
static void
dissect_isup_information_indicators_parameter(int offset, int len)
{
  int information_indicators;

  information_indicators = tvb_get_guint16(offset);

  proto_tree_add_text(2, "Calling party address response indicator: %s", val_to_str( (information_indicators & BA_16BIT_MASK) >> 8, isup_calling_party_address_response_ind_value, NULL));
  proto_tree_add_text(2, "Hold provided indicator: %s", tf_to_str( (information_indicators & C_16BIT_MASK) >> 10, isup_hold_provided_ind_value));
  proto_tree_add_text(2, "Calling party's category response indicator: %s", tf_to_str( (information_indicators & F_16BIT_MASK) >> 13, isup_calling_partys_category_response_ind_value));
  proto_tree_add_text(2, "Charge information response indicator: %s", tf_to_str( (information_indicators & G_16BIT_MASK) >> 14, isup_charge_information_response_ind_value));
  proto_tree_add_text(2, "Solicited indicator: %s", tf_to_str( (information_indicators & H_16BIT_MASK) >> 15, isup_solicited_information_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter Continuity Indicators
 */
static void
dissect_isup_continuity_indicators_parameter(int offset, int len)
{
  unsigned char continuity_indicators;

  continuity_indicators = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Continuity indicator: %s", tf_to_str( (continuity_indicators & A_8BIT_MASK) >> 0, isup_continuity_ind_value));
}
/* ------------------------------------------------------------------
 Dissector Parameter Backward Call Indicators
 */
static void
dissect_isup_backward_call_indicators_parameter(int offset, int len)
{
  int backward_call_ind;

  backward_call_ind = tvb_get_guint16(offset);

  proto_tree_add_text(2, "Charge indicator: %s", val_to_str( (backward_call_ind & BA_16BIT_MASK) >> 8, isup_charge_ind_value, NULL));
  proto_tree_add_text(2, "Called party's status indicator: %s", val_to_str( (backward_call_ind & DC_16BIT_MASK) >> 10, isup_called_partys_status_ind_value, NULL));
  proto_tree_add_text(2, "Called party's category indicator: %s", val_to_str( (backward_call_ind & FE_16BIT_MASK) >> 12, isup_called_partys_category_ind_value, NULL));
  proto_tree_add_text(2, "End-to-end method ind.: %s", val_to_str( (backward_call_ind & HG_16BIT_MASK) >> 14, isup_end_to_end_method_ind_value, NULL));
  proto_tree_add_text(2, "Interworking indicator: %s", tf_to_str( (backward_call_ind & I_16BIT_MASK) >> 0, isup_interworking_ind_value));
  proto_tree_add_text(2, "End-to-end information indicator: %s", tf_to_str( (backward_call_ind & J_16BIT_MASK) >> 1, isup_end_to_end_info_ind_value));
  proto_tree_add_text(2, "ISDN user part indicator: %s", tf_to_str( (backward_call_ind & K_16BIT_MASK) >> 2, isup_ISDN_user_part_ind_value));
  proto_tree_add_text(2, "Holding indicator: %s", tf_to_str( (backward_call_ind & L_16BIT_MASK) >> 3, isup_holding_ind_value));
  proto_tree_add_text(2, "ISDN access indicator: %s", tf_to_str( (backward_call_ind & M_16BIT_MASK) >> 4, isup_ISDN_terminating_access_ind_value));
  proto_tree_add_text(2, "Echo Control Device Indicator: %s", tf_to_str( (backward_call_ind & N_16BIT_MASK) >> 5, isup_echo_control_device_ind_value));
  proto_tree_add_text(2, "SCCP method indicator: %s", val_to_str( (backward_call_ind & PO_16BIT_MASK) >> 6, isup_SCCP_method_ind_value, NULL));

}
/* ------------------------------------------------------------------
  Dissector Parameter Cause Indicators - no detailed dissection since defined in Rec. Q.850
 */


static void
dissect_isup_cause_indicators_parameter(int offset, int len)
{ 

  dissect_q931_cause_ie(offset, len);

}

/* ------------------------------------------------------------------
  Dissector Parameter Suspend/Resume Indicators
 */
static void
dissect_isup_suspend_resume_indicators_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Suspend/Resume indicator: %s", tf_to_str( (indicators & A_8BIT_MASK) >> 0, isup_suspend_resume_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter Range and Status Indicators
 */
static void
dissect_isup_range_and_status_parameter(int offset, int len)
{
  unsigned char range;

  range = tvb_get_guint8(offset);

  proto_tree_add_text_i(2, "Range indicator: %d", range);

  len-=RANGE_LENGTH;
  offset+=RANGE_LENGTH;

  if (len > 0)
    proto_tree_add_text(2, "Status subfield:%s", tvb_bytes_to_BinaryStr(offset, len ));
  else
    proto_tree_add_text(2, "Status subfield is not present with this message type%s", "");
}
/* ------------------------------------------------------------------
  Dissector Parameter Circuit group supervision message type
 */
static void
dissect_isup_circuit_group_supervision_message_type_parameter(int offset, int len)
{
  unsigned char cgs_message_type;

  cgs_message_type = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Circuit group supervision message type: %s", val_to_str( (cgs_message_type & BA_8BIT_MASK) >> 0, isup_cgs_message_type_value, NULL));

}
/* ------------------------------------------------------------------
  Dissector Parameter Facility indicator parameter
 */
static void
dissect_isup_facility_ind_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text(2, "Facility indicator: %s", val_to_str( indicator, isup_facility_ind_value, "spare"));

}
/* ------------------------------------------------------------------
  Dissector Parameter Circuit state indicator
 */
static void
dissect_isup_circuit_state_ind_parameter(int offset, int len)
{
  unsigned char circuit_state;
  int i=0;

  while(len > 0)
  {
    proto_tree_add_text_i(2, "Circuit# CIC+%d state", i);
    circuit_state = tvb_get_guint8(offset);
    if ((circuit_state & DC_8BIT_MASK) == 0)
	{
	  proto_tree_add_text(3, "Maintenance blocking state: %s", val_to_str( (circuit_state & BA_8BIT_MASK) >> 0, isup_mtc_blocking_state_DC00_value, NULL));
	}
    else 
	{
  	  proto_tree_add_text(3, "Maintenance blocking state: %s", val_to_str( (circuit_state & BA_8BIT_MASK) >> 0, isup_mtc_blocking_state_DCnot00_value, NULL));
  	  proto_tree_add_text(3, "Call processing state: %s", val_to_str( (circuit_state & DC_8BIT_MASK) >> 2, isup_call_processing_state_value, NULL));
  	  proto_tree_add_text(3, "HW blocking state: %s", val_to_str( (circuit_state & FE_8BIT_MASK) >> 4, isup_HW_blocking_state_value, NULL));
    }
	i++;
    offset++;
    len--;
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Event information
 */
static void
dissect_isup_event_information_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Event indicator: %s", val_to_str( (indicators & GFEDCBA_8BIT_MASK) >> 0, isup_event_ind_value, "spare"));
  proto_tree_add_text(2, "Event presentation restricted indicator: %s", tf_to_str( (indicators & H_8BIT_MASK) >> 7, isup_event_presentation_restricted_ind_value));
}
/*
#define	Q931_PROTOCOL_DISCRIMINATOR_IA5		0x04
#define Q931_PROTOCOL_DISCRIMINATOR_ASN1	0x05

static const value_string q931_protocol_discriminator_vals[] = {
	{ 0x00,					"User-specific protocol" },
	{ 0x01,					"OSI high layer protocols" },
	{ 0x02,					"X.244" },
	{ Q931_PROTOCOL_DISCRIMINATOR_IA5,	"IA5 characters" },
	{ Q931_PROTOCOL_DISCRIMINATOR_ASN1,	"X.208 and X.209 coded user information" },
	{ 0x07,					"V.120 rate adaption" },
	{ 0x08,					"Q.931/I.451 user-network call control messages" },
	{ 0,					NULL }
};
*/

/* ------------------------------------------------------------------
  Dissector Parameter User-to-user information- no detailed dissection since defined in Rec. Q.931
 */
static void
dissect_isup_user_to_user_information_parameter(int offset, int len)
{ 

  proto_tree_add_text(2, "User-to-user info (-> Q.931)%s", "");

  unsigned int octet;

  if (len == 0)
		return;
  octet = tvb_get_guint8(offset);
  proto_tree_add_text(3, "Protocol discriminator: %s", val_to_str(octet, q931_protocol_discriminator_vals, "Unknown (0x%02x)"));
  offset += 1;
  len -= 1;

  if (len == 0)
 	return;
  switch (octet) 
  {
	case Q931_PROTOCOL_DISCRIMINATOR_IA5:
		proto_tree_add_text(3, "User information: 0x%s", tvb_bytes_to_str(offset, len));
		break;
	default:
		proto_tree_add_text(3, "User information: 0x%s", tvb_bytes_to_str(offset, len));
		break;
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Call Reference
 */
static void
dissect_isup_call_reference_parameter(int offset, int len)
{
  unsigned long call_id;
  unsigned int spc;

  call_id = tvb_get_guint24(offset);
  spc = tvb_get_guint16i( offset + CALL_ID_LENGTH ) & 0x3FFF; /*since 1st 2 bits spare */

  proto_tree_add_text_u(2, "Call identity: %u", call_id);
  proto_tree_add_text_i(2, "Signalling Point Code: %d", spc);
}
/* ------------------------------------------------------------------
  Dissector Parameter Access Transport - no detailed dissection since defined in Rec. Q.931
 */
static void
dissect_isup_access_transport_parameter(int offset, int len)
{ 
	dissect_q931_pdu(offset, len);
//	proto_tree_add_text(2, "Access transport parameter field (-> Q.931): 0x%s", tvb_bytes_to_str(offset, len));
}
/* ------------------------------------------------------------------
  Dissector Parameter Optional Forward Call indicators
 */
static void
dissect_isup_optional_forward_call_indicators_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Closed user group call indicator: %s", val_to_str( (indicators & BA_8BIT_MASK) >> 0, isup_CUG_call_ind_value, NULL));
  proto_tree_add_text(2, "Simple segmentation indicator: %s", tf_to_str( (indicators & C_8BIT_MASK) >> 2, isup_simple_segmentation_ind_value));
  proto_tree_add_text(2, "Connected line identity request indicator: %s", tf_to_str( (indicators & H_8BIT_MASK) >> 7, isup_connected_line_identity_request_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter calling party number
 */
static void
dissect_isup_calling_party_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "NI indicator: %s", tf_to_str( (indicators2 & ISUP_NI_MASK) , isup_NI_ind_value));
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));
  proto_tree_add_text(2, "Screening indicator: %s", val_to_str( (indicators2 & ISUP_SCREENING_IND_MASK) >> 0 , isup_screening_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Calling Party Number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Original called  number
 */
static void
dissect_isup_original_called_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Original Called Number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Redirecting number
 */
static void
dissect_isup_redirecting_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Redirecting Number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Redirection number
 */
static void
dissect_isup_redirection_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char called_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_called_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "INN indicator: %s", tf_to_str( (indicators2 & ISUP_INN_MASK) , isup_INN_ind_value));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    called_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      called_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  called_number[i++] = '\0';
  proto_tree_add_text(2, "Redirection Number: %s", called_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Connection request
 */
static void
dissect_isup_connection_request_parameter(int offset, int len)
{
  unsigned long local_ref;
  int spc;
  unsigned char protocol_class, credit;

  local_ref = tvb_get_guint24(offset);
  proto_tree_add_text_u(2, "Local Reference: %u", local_ref);
  offset += LOCAL_REF_LENGTH;
  spc = tvb_get_guint16i(offset) & 0x3FFF; /*since 1st 2 bits spare */
  proto_tree_add_text_i(2, "Signalling Point Code: %i", spc);
  offset += SPC_LENGTH;
  protocol_class = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Protocol Class: %d", protocol_class);
  offset += PROTOCOL_CLASS_LENGTH;
  credit = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Credit: %d", credit);
  offset += CREDIT_LENGTH;
}
/* ------------------------------------------------------------------
  Dissector Parameter Redirection information
 */
static void
dissect_isup_redirection_information_parameter(int offset, int len)
{
  if (len == 2){
    int indicators;
    indicators = tvb_get_guint16(offset);
    proto_tree_add_text(2, "Redirection indicator: %s", val_to_str( (indicators & CBA_16BIT_MASK) >> 8, isup_redirecting_ind_value, NULL));
    proto_tree_add_text(2, "Original redirection reason: %s", val_to_str( (indicators & HGFE_16BIT_MASK) >> 12, isup_original_redirection_reason_value, NULL));
    proto_tree_add_text_i(2, "Redirection counter: %d", (indicators & KJI_16BIT_MASK) >> 0);
    proto_tree_add_text(2, "Redirection reason: %s", val_to_str( (indicators & PONM_16BIT_MASK) >> 4, isup_redirection_reason_value, NULL));
  }
  else /* ISUP'88 (blue book) */
  { 
    int indicators;
    indicators = tvb_get_guint8(offset) * 0x100; /*since 2nd octet isn't present*/
    proto_tree_add_text(2, "Redirection indicator: %s", val_to_str( (indicators & CBA_16BIT_MASK) >> 8, isup_redirecting_ind_value, NULL));
    proto_tree_add_text(2, "Original redirection reason: %s", val_to_str( (indicators & HGFE_16BIT_MASK) >> 12, isup_original_redirection_reason_value, NULL));
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Closed user group interlock code
 */
static void
dissect_isup_closed_user_group_interlock_code_parameter(int offset, int len)
{
  char NI_digits[5]="";
  unsigned char digit_pair;
  unsigned int bin_code;

  digit_pair = tvb_get_guint8(offset++);
  NI_digits[0] = number_to_char((digit_pair & HGFE_8BIT_MASK) / 0x10);
  NI_digits[1] = number_to_char(digit_pair & DCBA_8BIT_MASK);
  digit_pair = tvb_get_guint8(offset++);
  NI_digits[2] = number_to_char((digit_pair & HGFE_8BIT_MASK) / 0x10);
  NI_digits[3] = number_to_char(digit_pair & DCBA_8BIT_MASK);
  NI_digits[4] = '\0';
  proto_tree_add_text(2, "Network Identity: %s", NI_digits);
  bin_code = tvb_get_guint16(offset);
  proto_tree_add_text_i(2, "Binary Code: 0x%x", bin_code);
}
/* ------------------------------------------------------------------
  Dissector Parameter User service information- no detailed dissection since defined in Rec. Q.931
 */
static void
dissect_isup_user_service_information_parameter(int offset, int len)
{
	dissect_q931_bearer_capability_ie(offset, len);
//	proto_tree_add_text(2, "User service information (-> Q.931 Bearer_capability): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Signalling point code
 */
static void
dissect_isup_signalling_point_code_parameter(int offset, int len)
{
  unsigned int spc;

  spc = tvb_get_guint16i(offset) & 0x3FFF; /*since 1st 2 bits spare */
  proto_tree_add_text_i(2, "Signalling Point Code: %d", spc);

}
/* ------------------------------------------------------------------
  Dissector Parameter Connected number
 */
static void
dissect_isup_connected_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));
  proto_tree_add_text(2, "Screening indicator: %s", val_to_str( (indicators2 & ISUP_SCREENING_IND_MASK) >> 0 , isup_screening_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Connected Number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Transit network selection
 */
static void
dissect_isup_transit_network_selection_parameter(int offset, int len)
{
  unsigned char indicators;
  unsigned char address_digit_pair=0;
  int i=0;
  char network_id[MAXLENGTH]="";

  indicators = tvb_get_guint8(offset++);

  proto_tree_add_text(2, "Network identification plan: %s", val_to_str( (indicators & DCBA_8BIT_MASK) >> 0 , isup_network_identification_plan_value, NULL));
  proto_tree_add_text(2, "Type of network identification: %s", val_to_str( (indicators & GFE_8BIT_MASK) >> 4 , isup_type_of_network_identification_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  len -= 1;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    network_id[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      network_id[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      network_id[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  network_id[i++] = '\0';
  proto_tree_add_text(2, "Network identification: %s", network_id);
}
/* ------------------------------------------------------------------
  Dissector Parameter Circuit assignment map
 */
static void
dissect_isup_circuit_assignment_map_parameter(int offset, int len)
{ unsigned char map_type;

  map_type = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Map Type: %s", val_to_str( (map_type & FEDCBA_8BIT_MASK) >> 0 , isup_map_type_value, NULL));

  len-=1;
  offset+=1;

  if (len > 0)
    proto_tree_add_text(2, "Circuit assignment map (bit position indicates usage of corresponding circuit->3.69/Q.763): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Automatic congestion level
 */
static void
dissect_isup_automatic_congestion_level_parameter(int offset, int len)
{ 
  unsigned char congestion_level;

  congestion_level = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Automatic congestion level: %s", val_to_str( congestion_level , isup_auto_congestion_level_value, NULL));

}
/* ------------------------------------------------------------------
  Dissector Parameter Optional backward Call indicators
 */
static void
dissect_isup_optional_backward_call_indicators_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);

  proto_tree_add_text(2, "In-band information indicator: %s", tf_to_str( (indicators & A_8BIT_MASK) , isup_inband_information_ind_value));
  proto_tree_add_text(2, "Call diversion may occur indicator: %s", tf_to_str( (indicators & B_8BIT_MASK) >> 1, isup_call_diversion_may_occur_ind_value));
  proto_tree_add_text(2, "Simple segmentation indicator: %s", tf_to_str( (indicators & C_8BIT_MASK) >> 2, isup_simple_segmentation_ind_value));
  proto_tree_add_text(2, "MLPP user indicator: %s", tf_to_str( (indicators & D_8BIT_MASK) >> 3, isup_MLPP_user_ind_value));

}
/* ------------------------------------------------------------------
  Dissector Parameter User-to-user indicators
 */
static void
dissect_isup_user_to_user_indicators_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);

  proto_tree_add_text_i(2, "User-to-user indicators: 0x%x (refer to 3.60/Q.763 for detailed decoding)", indicators );

}
/* ------------------------------------------------------------------
  Dissector Parameter Original ISC point code
 */
static void
dissect_isup_original_isc_point_code_parameter(int offset, int len)
{
  unsigned int spc;

  spc = tvb_get_guint16i(offset) & 0x3FFF; /*since 1st 2 bits spare */
  proto_tree_add_text_i(2, "Origination ISC point code: %d", spc);
}
/* ------------------------------------------------------------------
  Dissector Parameter Generic notification indicator
 */
static void
dissect_isup_generic_notification_indicator_parameter(int offset, int len)
{
  unsigned char indicators;

  indicators = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Generic notification indicator: 0x%x (refer to 3.25/Q.763 for detailed decoding)", indicators );
}
/* ------------------------------------------------------------------
  Dissector Parameter Call history information
 */
static void
dissect_isup_call_history_information_parameter(int offset, int len)
{
  unsigned int info;

  info = tvb_get_guint16(offset);
  proto_tree_add_text_i(2, "Call history info: propagation delay = %d ms", info );
}
/* ------------------------------------------------------------------
  Dissector Parameter Access delivery information
 */
static void
dissect_isup_access_delivery_information_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Access delivery indicator: %s", tf_to_str( (indicator & A_8BIT_MASK) , isup_access_delivery_ind_value));

}
/* ------------------------------------------------------------------
  Dissector Parameter Network specific facility
 */
static void
dissect_isup_network_specific_facility_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Network specific facility (refer to 3.36/Q.763 for detailed decoding): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter User service information prime
 */
static void
dissect_isup_user_service_information_prime_parameter(int offset, int len)
{ 
  dissect_q931_bearer_capability_ie(offset, len);
//  proto_tree_add_text(2, "User service information prime (-> Q.931): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Propagation delay counter
 */
static void
dissect_isup_propagation_delay_counter_parameter(int offset, int len)
{
  unsigned int info;

  info = tvb_get_guint16(offset);

  proto_tree_add_text_i(2, "Propagation delay counter = %d ms", info );
}
/* ------------------------------------------------------------------
  Dissector Parameter Remote operations
 */
static void
dissect_isup_remote_operations_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Remote operations: 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Service activation
 */
static void
dissect_isup_service_activation_parameter(int offset, int len)
{
  int i;
  unsigned char feature_code;
  static char str_parameter[128];
  for (i=0; i< len; i++)
  {
    feature_code = tvb_get_guint8(offset++);
	sprintf(&str_parameter[0], "Feature Code %u: %u", i+1, feature_code );
	proto_tree_add_text(2, "%s", &str_parameter[0]);
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter User service information prime - no detailed dissection since defined in Rec. Q.931
 */
static void
dissect_isup_user_teleservice_information_parameter(int offset, int len)
{
	dissect_q931_high_layer_compat_ie(offset, len);
//  proto_tree_add_text(2, "User teleservice information (-> Q.931): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Transmission medium requirement used
 */
static void
dissect_isup_transmission_medium_used_parameter(int offset, int len)
{
  unsigned char transmission_medium_requirement;

  transmission_medium_requirement = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Transmission medium used: %s", val_to_str( transmission_medium_requirement , isup_transmission_medium_requirement_prime_value, NULL));

}
/* ------------------------------------------------------------------
  Dissector Parameter Call diversion information
 */
static void
dissect_isup_call_diversion_information_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Call diversion information: 0x%x (refer to 3.6/Q.763 for detailed decoding)", indicator);

}
/* ------------------------------------------------------------------
  Dissector Parameter Echo control  information
 */
static void
dissect_isup_echo_control_information_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Echo control information: 0x%x (refer to 3.19/Q.763 for detailed decoding)", indicator);
}
/* ------------------------------------------------------------------
  Dissector Parameter Message compatibility information
 */
static void
dissect_isup_message_compatibility_information_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Message compatibility information (refer to 3.33/Q.763 for detailed decoding): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter compatibility information
 */
static void
dissect_isup_parameter_compatibility_information_parameter(int offset, int len)
{
  proto_tree_add_text(2, "Parameter compatibility information (refer to 3.41/Q.763 for detailed decoding): 0x%s", tvb_bytes_to_str(offset, len ));

}
/* ------------------------------------------------------------------
  Dissector Parameter MLPP precedence
 */
static void
dissect_isup_mlpp_precedence_parameter(int offset, int len)
{
  char NI_digits[5]="";
  unsigned char indicators, digit_pair;
  unsigned long bin_code;

  indicators = tvb_get_guint8(offset++);
  proto_tree_add_text_i(2, "LFB (Bits 6+7) and precedence level (Bits 1-4): 0x%x", indicators);

  digit_pair = tvb_get_guint8(offset++);
  NI_digits[0] = number_to_char((digit_pair & HGFE_8BIT_MASK) / 0x10);
  NI_digits[1] = number_to_char(digit_pair & DCBA_8BIT_MASK);
  digit_pair = tvb_get_guint8(offset++);
  NI_digits[2] = number_to_char((digit_pair & HGFE_8BIT_MASK) / 0x10);
  NI_digits[3] = number_to_char(digit_pair & DCBA_8BIT_MASK);
  NI_digits[4] = '\0';
  proto_tree_add_text(2, "Network Identity: %s", NI_digits);
  bin_code = tvb_get_guint24(offset);
  proto_tree_add_text_u(2, "MLPP service domain: 0x%x", bin_code);
}
/* ------------------------------------------------------------------
  Dissector Parameter MCID request indicators
 */
static void
dissect_isup_mcid_request_indicators_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "MCID request indicators: 0x%x (MCID requested by Bit1=1, Holding requested by Bit2=1 see 3.31/Q.763)", indicator);
}
/* ------------------------------------------------------------------
  Dissector Parameter MCID response indicators
 */
static void
dissect_isup_mcid_response_indicators_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "MCID response indicators: 0x%x (MCID included if Bit1=1, Holding provided if Bit2=1 see 3.32/Q.763)", indicator);
}
/* ------------------------------------------------------------------
  Dissector Parameter Hop counter
 */
static void
dissect_isup_hop_counter_parameter(int offset, int len)
{
  unsigned char counter;

  counter = tvb_get_guint8(offset) & EDCBA_8BIT_MASK; /* since bits H,G and F are spare */
  proto_tree_add_text_i(2, "Hop counter: %d", counter);
}
/* ------------------------------------------------------------------
  Dissector Parameter Transmission medium requirement prime
 */
static void
dissect_isup_transmission_medium_requirement_prime_parameter(int offset, int len)
{
  unsigned char transmission_medium_requirement;

  transmission_medium_requirement = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Transmission medium requirement prime: %s", val_to_str( transmission_medium_requirement , isup_transmission_medium_requirement_prime_value, NULL));

}

/* ------------------------------------------------------------------
  Dissector Parameter location number
 */
static void
dissect_isup_location_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "INN indicator: %s", tf_to_str( (indicators2 & ISUP_INN_MASK) , isup_INN_ind_value));
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  if ((indicators2 & GFE_8BIT_MASK) == 0x50)
    proto_tree_add_text(2, "Different meaning for Location Number: Numbering plan indicator = private numbering plan%s", "");
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));
  proto_tree_add_text(2, "Screening indicator: %s", val_to_str( (indicators2 & ISUP_SCREENING_IND_MASK) >> 0 , isup_screening_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Location number: %s", calling_number);

}
/* ------------------------------------------------------------------
  Dissector Parameter Redirection number restiriction
 */
static void
dissect_isup_redirection_number_restriction_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  switch (indicator & BA_8BIT_MASK) {
  case 0:
    proto_tree_add_text(2, "Presentation indicator: Presentation allowed%s", "");
    break;
  case 1:
    proto_tree_add_text(2, "Presentation indicator: Presentation restricted%s", "");
    break;
  default:
    proto_tree_add_text(2, "Presentation indicator: spare%s", "");
    break;
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Call transfer identity
 */
static void
dissect_isup_call_transfer_reference_parameter(int offset, int len)
{ 
  unsigned char id;

  id = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "Call transfer identity: %d", id);
 }
/* ------------------------------------------------------------------
  Dissector Parameter Loop prevention
 */
static void
dissect_isup_loop_prevention_indicators_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  if ((indicator & A_8BIT_MASK)==0) 
  {
    proto_tree_add_text(2, "Type: Request%s", "");
  }
  else 
  {
    proto_tree_add_text(2, "Type: Response%s", "");
    proto_tree_add_text(2, "Response indicator: %s", val_to_str( (indicator & CB_8BIT_MASK) >> 1 , isup_loop_prevention_response_ind_value, NULL));
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Call transfer number
 */
static void
dissect_isup_call_transfer_number_parameter(int offset, int len)
{  
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  if ((indicators2 & GFE_8BIT_MASK) == 0x50)
    proto_tree_add_text(2, "Different meaning for Call Transfer Number: Numbering plan indicator = private numbering plan%s", "");
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));
  proto_tree_add_text(2, "Screening indicator: %s", val_to_str( (indicators2 & ISUP_SCREENING_IND_MASK) >> 0 , isup_screening_ind_enhanced_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Call transfer number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter CCSS
 */
static void
dissect_isup_ccss_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  if ((indicator & A_8BIT_MASK)==0) 
  {
    proto_tree_add_text(2, "CCSS call indicator: no indication%s", "");
  }
  else 
  {
    proto_tree_add_text(2, "CCSS call indicator: CCSS call%s", "");
  }
}
/* ------------------------------------------------------------------
 Parameter Forward GVNS
 */
static void
dissect_isup_forward_gvns_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Forward GVNS (refer to 3.66/Q.763 for detailed decoding): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
 Parameter Redirect capability
 */
static void
dissect_isup_redirect_capability_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Redirect capability (format is a national matter): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Backward GVNS
 */
static void
dissect_isup_backward_gvns_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);

  proto_tree_add_text_i(2, "Backward GVNS: 0x%x (refer to 3.62/Q.763 for detailed decoding): 0x%s", indicator);
}
/* ------------------------------------------------------------------
  Dissector Parameter Network management controls
 */
static void
dissect_isup_network_management_controls_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Temporary alternative routing indicator: %s", tf_to_str( (indicator & A_8BIT_MASK) >> 0, isup_temporary_alternative_routing_ind_value));
  proto_tree_add_text(2, "Extension indicator: %s", tf_to_str( (indicator & H_8BIT_MASK) >> 7, isup_extension_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter Correlation id - no detailed dissection since defined in Rec. Q.1281
 */
static void
dissect_isup_correlation_id_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Correlation ID (-> Q.1281): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter SCF id - no detailed dissection since defined in Rec. Q.1281
 */
static void
dissect_isup_scf_id_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "SCF ID (-> Q.1281): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Call diversion treatment indicators
 */
static void
dissect_isup_call_diversion_treatment_indicators_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text(2, "Call to be diverted indicator: %s", val_to_str( (indicator & BA_8BIT_MASK) >> 0 , isup_call_to_be_diverted_ind_value, NULL));
  proto_tree_add_text(2, "Extension indicator: %s", tf_to_str( (indicator & H_8BIT_MASK) >> 7, isup_extension_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter called IN  number
 */
static void
dissect_isup_called_in_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));

  len -= 2;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Called IN Number: %s", calling_number);
}
/* ------------------------------------------------------------------
  Dissector Parameter Call offering treatment indicators
 */
static void
dissect_isup_call_offering_treatment_indicators_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Call to be offered indicator: %s", val_to_str( (indicator & BA_8BIT_MASK) >> 0 , isup_call_to_be_offered_ind_value, NULL));
  proto_tree_add_text(2, "Extension indicator: %s", tf_to_str( (indicator & H_8BIT_MASK) >> 7, isup_extension_ind_value));
}
/* ------------------------------------------------------------------
 Parameter Charged party identification
 */
static void
dissect_isup_charged_party_identification_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Charged party identification (format is national network specific): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Conference treatment indicators
 */
static void
dissect_isup_conference_treatment_indicators_parameter(int offset, int len)
{ 
  unsigned char indicator;
  indicator = tvb_get_guint8(offset);

  proto_tree_add_text(2, "Conference acceptance indicator: %s", val_to_str( (indicator & BA_8BIT_MASK) >> 0 , isup_conference_acceptance_ind_value, NULL));
  proto_tree_add_text(2, "Extension indicator: %s", tf_to_str( (indicator & H_8BIT_MASK) >> 7, isup_extension_ind_value));
}
/* ------------------------------------------------------------------
  Dissector Parameter Display information
 */
static void
dissect_isup_display_information_parameter(int offset, int len)
{ 
  dissect_q931_ia5_ie(offset, len, "Display information");
 // proto_tree_add_text(2, "Display information (-> Q.931): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
 Parameter UID action indicators
 */
static void
dissect_isup_uid_action_indicators_parameter(int offset, int len)
{ 
  unsigned char indicator;
  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "UID action indicators: 0x%x (refer to 3.78/Q.763 for detailed decoding)", indicator);
}
/* ------------------------------------------------------------------
 Parameter UID capability indicators
 */
static void
dissect_isup_uid_capability_indicators_parameter(int offset, int len)
{
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  proto_tree_add_text_i(2, "UID capability indicators: 0x%x (refer to 3.79/Q.763 for detailed decoding)", indicator);
}
/* ------------------------------------------------------------------
 Parameter Redirect counter
 */
static void
dissect_isup_redirect_counter_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Redirect counter (format is a national matter): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------
  Dissector Parameter Collect call request
 */
static void
dissect_isup_collect_call_request_parameter(int offset, int len)
{ 
  unsigned char indicator;

  indicator = tvb_get_guint8(offset);
  if ((indicator & A_8BIT_MASK) == 0) 
  {
    proto_tree_add_text(2, "Collect call request indicator: no indication%s", "");
  }
  else 
  {
    proto_tree_add_text(2, "Collect call request indicator: collect call requested%s", "");
  }
}
/* ------------------------------------------------------------------
  Dissector Parameter Generic number
 */
static void
dissect_isup_generic_number_parameter(int offset, int len)
{
  unsigned char indicators1, indicators2, nr_qualifier_ind;
  unsigned char address_digit_pair=0;
  int i=0;
  char calling_number[MAXLENGTH]="";

  nr_qualifier_ind = tvb_get_guint8(offset++);
  proto_tree_add_text_i(2, "Number qualifier indicator: 0x%x (refer to 3.26/Q.763 for detailed decoding)", nr_qualifier_ind);

  indicators1 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "Nature of address indicator: %s", val_to_str( (indicators1 & ISUP_NATURE_OF_ADDRESS_IND_MASK) , isup_calling_party_nature_of_address_ind_value, NULL));
  proto_tree_add_text(2, "Odd/even indicator: %s", tf_to_str( (indicators1 & ISUP_ODD_EVEN_MASK) , isup_odd_even_ind_value));

  indicators2 = tvb_get_guint8(offset++);
  proto_tree_add_text(2, "NI indicator: %s", tf_to_str( (indicators2 & ISUP_NI_MASK) , isup_NI_ind_value));
  proto_tree_add_text(2, "Numbering plan indicator: %s", val_to_str( (indicators2 & ISUP_NUMBERING_PLAN_IND_MASK) >>4 , isup_numbering_plan_ind_value, NULL));
  if ((indicators2 & GFE_8BIT_MASK) == 0x50)
    proto_tree_add_text(2, "Different meaning for Generic Number: Numbering plan indicator = private numbering plan%s", "");
  proto_tree_add_text(2, "Address presentation restricted indicator: %s", val_to_str( (indicators2 & ISUP_ADDRESS_PRESENTATION_RESTR_IND_MASK) >>2 , isup_address_presentation_restricted_ind_value, NULL));
  proto_tree_add_text(2, "Screening indicator: %s", val_to_str( (indicators2 & ISUP_SCREENING_IND_MASK) >> 0 , isup_screening_ind_enhanced_value, NULL));

  len -= 3;
  while(len > 0){
    address_digit_pair = tvb_get_guint8(offset++);
    calling_number[i++] = number_to_char(address_digit_pair & ISUP_ODD_ADDRESS_SIGNAL_DIGIT_MASK);
    if ((len - 1) > 0 ){
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4);
    }
    len--;
  }

  if  ( (indicators1 & 0x80) == 0 ){ /* Even Indicator set -> last even digit is valid & has be displayed */
      calling_number[i++] = number_to_char((address_digit_pair & ISUP_EVEN_ADDRESS_SIGNAL_DIGIT_MASK) >> 4 );
  }
  calling_number[i++] = '\0';
  proto_tree_add_text(2, "Generic number: %s", calling_number);
}
/* ------------------------------------------------------------------
 Dissector Parameter Generic digits
 */
static void
dissect_isup_generic_digits_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Generic digits (refer to 3.24/Q.673 for detailed decoding): 0x%s", tvb_bytes_to_str(offset, len ));
}
/* ------------------------------------------------------------------ */
static void
dissect_isup_unknown_parameter(int offset, int len)
{ 
  proto_tree_add_text(2, "Parameter Type unknown/reserved: 0x%s", tvb_bytes_to_str(offset-1, len+1 ));
}
/* ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
  Dissector all optional parameters
*/
static void
dissect_isup_optional_parameter(int offset, int len)
{ 
  int parameter_type, parameter_length, actual_length;

  /* Dissect all optional parameters while end of message isn't reached */
  parameter_type = 0xFF; /* Start-initializiation since parameter_type is used for while-condition */

  while ( ((len-offset) >= 1) && (parameter_type != PARAM_TYPE_END_OF_OPT_PARAMS))
  {
    parameter_type = tvb_get_guint8(offset);

    if (parameter_type != PARAM_TYPE_END_OF_OPT_PARAMS)
	{
	  offset += PARAMETER_TYPE_LENGTH;
      parameter_length = tvb_get_guint8(offset);
	  offset += PARAMETER_LENGTH_IND_LENGTH;

	  print_parameter_title(offset, parameter_length, parameter_type);
	 
      actual_length = len - offset;
      if (actual_length > 0)
	  {
		switch (parameter_type) 
		{
			case PARAM_TYPE_CALL_REF:
			dissect_isup_call_reference_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_TRANSM_MEDIUM_REQU:
			dissect_isup_transmission_medium_requirement_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_ACC_TRANSP:
			dissect_isup_access_transport_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALLED_PARTY_NR:
			dissect_isup_called_party_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_SUBSQT_NR:
			dissect_isup_subsequent_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_NATURE_OF_CONN_IND:
			dissect_isup_nature_of_connection_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_FORW_CALL_IND:
			dissect_isup_forward_call_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_OPT_FORW_CALL_IND:
			dissect_isup_optional_forward_call_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALLING_PRTY_CATEG:
			dissect_isup_calling_partys_category_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALLING_PARTY_NR:
			dissect_isup_calling_party_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIRECTING_NR:
			dissect_isup_redirecting_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIRECTION_NR:
			dissect_isup_redirection_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CONNECTION_REQ:
			dissect_isup_connection_request_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_INFO_REQ_IND:
			dissect_isup_information_request_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_INFO_IND:
			dissect_isup_information_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CONTINUITY_IND:
			dissect_isup_continuity_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_BACKW_CALL_IND:
			dissect_isup_backward_call_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CAUSE_INDICATORS:
			dissect_isup_cause_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIRECTION_INFO:
			dissect_isup_redirection_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CIRC_GRP_SV_MSG_TYPE:
			dissect_isup_circuit_group_supervision_message_type_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_RANGE_AND_STATUS:
			dissect_isup_range_and_status_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_FACILITY_IND:
			dissect_isup_facility_ind_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CLSD_USR_GRP_ILOCK_CD:
			dissect_isup_closed_user_group_interlock_code_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_USER_SERVICE_INFO:
			dissect_isup_user_service_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_SIGNALLING_POINT_CODE:
			dissect_isup_signalling_point_code_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_USER_TO_USER_INFO:
			dissect_isup_user_to_user_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CONNECTED_NR:
			dissect_isup_connected_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_SUSP_RESUME_IND:
			dissect_isup_suspend_resume_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_TRANSIT_NETW_SELECT:
			dissect_isup_transit_network_selection_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_EVENT_INFO:
			dissect_isup_event_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CIRC_ASSIGN_MAP:
			dissect_isup_circuit_assignment_map_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CIRC_STATE_IND:
			dissect_isup_circuit_state_ind_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_AUTO_CONG_LEVEL:
			dissect_isup_automatic_congestion_level_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_ORIG_CALLED_NR:
			dissect_isup_original_called_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_OPT_BACKW_CALL_IND:
			dissect_isup_optional_backward_call_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_USER_TO_USER_IND:
			dissect_isup_user_to_user_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_ORIG_ISC_POINT_CODE:
			dissect_isup_original_isc_point_code_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_GENERIC_NOTIF_IND:
			dissect_isup_generic_notification_indicator_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_HIST_INFO :
			dissect_isup_call_history_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_ACC_DELIV_INFO:
			dissect_isup_access_delivery_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_NETW_SPECIFIC_FACLTY:
			dissect_isup_network_specific_facility_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_USER_SERVICE_INFO_PR:
			dissect_isup_user_service_information_prime_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_PROPAG_DELAY_COUNTER:
			dissect_isup_propagation_delay_counter_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REMOTE_OPERATIONS:
			dissect_isup_remote_operations_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_SERVICE_ACTIVATION:
			dissect_isup_service_activation_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_USER_TELESERV_INFO:
			dissect_isup_user_teleservice_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_TRANSM_MEDIUM_USED:
			dissect_isup_transmission_medium_used_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_DIV_INFO:
			dissect_isup_call_diversion_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_ECHO_CTRL_INFO:
			dissect_isup_echo_control_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_MSG_COMPAT_INFO:
			dissect_isup_message_compatibility_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_PARAM_COMPAT_INFO:
			dissect_isup_parameter_compatibility_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_MLPP_PRECEDENCE:
			dissect_isup_mlpp_precedence_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_MCID_REQ_IND:
			dissect_isup_mcid_request_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_MCID_RSP_IND:
			dissect_isup_mcid_response_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_HOP_COUNTER:
			dissect_isup_hop_counter_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_TRANSM_MEDIUM_RQUR_PR:
			dissect_isup_transmission_medium_requirement_prime_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_LOCATION_NR:
			dissect_isup_location_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIR_NR_RSTRCT:
			dissect_isup_redirection_number_restriction_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_TRANS_REF:
			dissect_isup_call_transfer_reference_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_LOOP_PREV_IND:
			dissect_isup_loop_prevention_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_TRANS_NR:
			dissect_isup_call_transfer_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CCSS:
			dissect_isup_ccss_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_FORW_GVNS:
			dissect_isup_forward_gvns_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_BACKW_GVNS:
			dissect_isup_backward_gvns_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIRECT_CAPAB:
			dissect_isup_redirect_capability_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_NETW_MGMT_CTRL:
			dissect_isup_network_management_controls_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CORRELATION_ID:
			dissect_isup_correlation_id_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_SCF_ID:
			dissect_isup_scf_id_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_DIV_TREAT_IND:
			dissect_isup_call_diversion_treatment_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALLED_IN_NR:
			dissect_isup_called_in_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CALL_OFF_TREAT_IND:
			dissect_isup_call_offering_treatment_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CHARGED_PARTY_IDENT:
			dissect_isup_charged_party_identification_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_CONF_TREAT_IND:
			dissect_isup_conference_treatment_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_DISPLAY_INFO:
			dissect_isup_display_information_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_UID_ACTION_IND:
			dissect_isup_uid_action_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_UID_CAPAB_IND:
			dissect_isup_uid_capability_indicators_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_REDIRECT_COUNTER:
			dissect_isup_redirect_counter_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_COLLECT_CALL_REQ:
			dissect_isup_collect_call_request_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_GENERIC_NR:
			dissect_isup_generic_number_parameter(offset,parameter_length);
			break;
			case PARAM_TYPE_GENERIC_DIGITS:
			dissect_isup_generic_digits_parameter(offset,parameter_length);
			break;
			default:
			dissect_isup_unknown_parameter(offset,parameter_length);
			break;
		}
		offset += parameter_length;
	  }
	}
    else 
	{
	  /* End of optional parameters is reached */
    }
  }
}

/* ------------------------------------------------------------------ */
/* Dissectors for all used message types                              */
/* Called by dissect_isup_message(),                                  */
/* call parameter dissectors in order of mandatory parameters         */
/* (since not labeled)                                                */
/* ------------------------------------------------------------------
  Dissector Message Type Initial address message
 */
int
dissect_isup_initial_address_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for first mandatory fixed parameter: Nature of Connection Indicators */
  parameter_type = PARAM_TYPE_NATURE_OF_CONN_IND;
  print_parameter_title(offset, NATURE_OF_CONNECTION_IND_LENGTH, parameter_type);
  actual_length = len;
  dissect_isup_nature_of_connection_indicators_parameter(offset,NATURE_OF_CONNECTION_IND_LENGTH);
  offset += NATURE_OF_CONNECTION_IND_LENGTH;

  /* Do stuff for 2nd mandatory fixed parameter: Forward Call Indicators */
  parameter_type =  PARAM_TYPE_FORW_CALL_IND;
  print_parameter_title(offset, FORWARD_CALL_IND_LENGTH, parameter_type);
  dissect_isup_forward_call_indicators_parameter(offset, FORWARD_CALL_IND_LENGTH);
  offset +=  FORWARD_CALL_IND_LENGTH;

  /* Do stuff for 3nd mandatory fixed parameter: Calling party's category */
  parameter_type = PARAM_TYPE_CALLING_PRTY_CATEG;
  print_parameter_title(offset, CALLING_PRTYS_CATEGORY_LENGTH, parameter_type);
  dissect_isup_calling_partys_category_parameter(offset, CALLING_PRTYS_CATEGORY_LENGTH);
  offset += CALLING_PRTYS_CATEGORY_LENGTH;

  /* Do stuff for 4th mandatory fixed parameter: Transmission medium requirement */
  parameter_type = PARAM_TYPE_TRANSM_MEDIUM_REQU;
  print_parameter_title(offset, TRANSMISSION_MEDIUM_REQUIREMENT_LENGTH, parameter_type);
  dissect_isup_transmission_medium_requirement_parameter(offset, TRANSMISSION_MEDIUM_REQUIREMENT_LENGTH);
  offset += TRANSMISSION_MEDIUM_REQUIREMENT_LENGTH;

  /* Do stuff for mandatory variable parameter Called party number */
  parameter_type = PARAM_TYPE_CALLED_PARTY_NR;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);

  dissect_isup_called_party_number_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type subsequent address message
 */
int dissect_isup_subsequent_address_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter Subsequent number */
  parameter_type = PARAM_TYPE_SUBSQT_NR;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);

  dissect_isup_subsequent_number_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}

/* ------------------------------------------------------------------
  Dissector Message Type Information request message
 */
int
dissect_isup_information_request_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: Information request indicators*/
  parameter_type = PARAM_TYPE_INFO_REQ_IND;
  print_parameter_title(offset, INFO_REQUEST_IND_LENGTH, parameter_type);
  dissect_isup_information_request_indicators_parameter(offset, INFO_REQUEST_IND_LENGTH);
  offset += INFO_REQUEST_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Information
 */
int
dissect_isup_information_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: Information  indicators*/
  parameter_type = PARAM_TYPE_INFO_IND;
  print_parameter_title(offset, INFO_IND_LENGTH, parameter_type);
  dissect_isup_information_indicators_parameter(offset, INFO_IND_LENGTH);
  offset += INFO_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Continuity
 */
int
dissect_isup_continuity_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: Continuity indicators*/
  parameter_type = PARAM_TYPE_CONTINUITY_IND;
  print_parameter_title(offset, CONTINUITY_IND_LENGTH, parameter_type);
  dissect_isup_continuity_indicators_parameter(offset, CONTINUITY_IND_LENGTH);
  offset += CONTINUITY_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Address complete
 */
int
dissect_isup_address_complete_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: backward call indicators*/
  parameter_type = PARAM_TYPE_BACKW_CALL_IND;
  print_parameter_title(offset, BACKWARD_CALL_IND_LENGTH, parameter_type);
  dissect_isup_backward_call_indicators_parameter(offset, BACKWARD_CALL_IND_LENGTH);
  offset += BACKWARD_CALL_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Connect
 */
int
dissect_isup_connect_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: backward call indicators*/
  parameter_type = PARAM_TYPE_BACKW_CALL_IND;
  print_parameter_title(offset, BACKWARD_CALL_IND_LENGTH, parameter_type);
  dissect_isup_backward_call_indicators_parameter(offset, BACKWARD_CALL_IND_LENGTH);
  offset += BACKWARD_CALL_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type release message
 */
int
dissect_isup_release_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter Cause indicators */
  parameter_type = PARAM_TYPE_CAUSE_INDICATORS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_cause_indicators_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Resume/Suspend
 */
int
dissect_isup_suspend_resume_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: backward call indicators*/
  parameter_type = PARAM_TYPE_SUSP_RESUME_IND;
  print_parameter_title(offset, SUSPEND_RESUME_IND_LENGTH, parameter_type);
  dissect_isup_suspend_resume_indicators_parameter(offset, SUSPEND_RESUME_IND_LENGTH);
  offset += SUSPEND_RESUME_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Circuit group reset/query message
 */
int
dissect_isup_circuit_group_reset_query_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter range and status*/
  parameter_type =  PARAM_TYPE_RANGE_AND_STATUS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_range_and_status_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Circuit group blocking/blocking ack/unblocking/unblocking ack messages
 */
int
dissect_isup_circuit_group_blocking_messages(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

   /* Do stuff for first mandatory fixed parameter: circuit group supervision message type*/
  parameter_type = PARAM_TYPE_CIRC_GRP_SV_MSG_TYPE;
  print_parameter_title(offset, CIRC_GRP_SV_MSG_TYPE_LENGTH, parameter_type);
  dissect_isup_circuit_group_supervision_message_type_parameter(offset, CIRC_GRP_SV_MSG_TYPE_LENGTH);
  offset += CIRC_GRP_SV_MSG_TYPE_LENGTH;

  /* Do stuff for mandatory variable parameter range and status*/
  parameter_type =  PARAM_TYPE_RANGE_AND_STATUS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_range_and_status_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}

/* ------------------------------------------------------------------
  Dissector Message Type Facility request/accepted
 */
int
dissect_isup_facility_request_accepted_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: facility indicators*/
  parameter_type = PARAM_TYPE_FACILITY_IND;
  print_parameter_title(offset, FACILITY_IND_LENGTH, parameter_type);
  dissect_isup_facility_ind_parameter(offset, FACILITY_IND_LENGTH);
  offset += FACILITY_IND_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Facility reject
 */
int
dissect_isup_facility_reject_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for first mandatory fixed parameter: facility indicators*/
  parameter_type = PARAM_TYPE_FACILITY_IND;
  print_parameter_title(offset, FACILITY_IND_LENGTH, parameter_type);
  dissect_isup_facility_ind_parameter(offset, FACILITY_IND_LENGTH);
  offset += FACILITY_IND_LENGTH;

  /* Do stuff for mandatory variable parameter Cause indicators */
  parameter_type =  PARAM_TYPE_CAUSE_INDICATORS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_cause_indicators_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Circuit group reset acknowledgement message
 */
int
dissect_isup_circuit_group_reset_acknowledgement_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter range and status*/
  parameter_type =  PARAM_TYPE_RANGE_AND_STATUS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_range_and_status_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Circuit group query response message
 */
int
dissect_isup_circuit_group_query_response_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for 1. mandatory variable parameter range and status*/
  parameter_type =  PARAM_TYPE_RANGE_AND_STATUS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_range_and_status_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  /* Do stuff for 2. mandatory variable parameter Circuit state indicator*/
  parameter_type = PARAM_TYPE_CIRC_STATE_IND;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_circuit_state_ind_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Call Progress
*/
int
dissect_isup_call_progress_message(int offset, int len)
{ 
  int parameter_type, actual_length;

  /* Do stuff for first mandatory fixed parameter: Event information*/
  parameter_type = PARAM_TYPE_EVENT_INFO;
  print_parameter_title(offset, EVENT_INFO_LENGTH, parameter_type);
  dissect_isup_event_information_parameter(offset, EVENT_INFO_LENGTH);
  offset += EVENT_INFO_LENGTH;
  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type User-to-User information
 */
int
dissect_isup_user_to_user_information_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter User-to-user information*/
  parameter_type =  PARAM_TYPE_USER_TO_USER_INFO;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_user_to_user_information_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}
/* ------------------------------------------------------------------
  Dissector Message Type Confusion
 */
int
dissect_isup_confusion_message(int offset, int len)
{ 
  int parameter_type, parameter_pointer, parameter_length, actual_length;

  /* Do stuff for mandatory variable parameter Cause indicators */
  parameter_type =  PARAM_TYPE_CAUSE_INDICATORS;
  parameter_pointer = tvb_get_guint8(offset);
  parameter_length = tvb_get_guint8(offset + parameter_pointer);
  print_parameter_title(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length, parameter_type);
  dissect_isup_cause_indicators_parameter(offset + parameter_pointer + PARAMETER_POINTER_LENGTH, parameter_length);
  offset += PARAMETER_POINTER_LENGTH;

  return offset;
}



/* ------------------------------------------------------------------ */
static void
dissect_isup_message(int len)
{
  int offset, bufferlength;
  unsigned char message_type, opt_parameter_pointer;
  int opt_part_possible = 0; /* default setting - for message types allowing optional
				     params explicitely set to TRUE in case statement */
  offset = CIC_OFFSET + CIC_LENGTH;

  /* Extract message type field */
  message_type = tvb_get_guint8(offset++);

   bufferlength = len;

   /* distinguish between message types:*/
   switch (message_type) {
     case MESSAGE_TYPE_INITIAL_ADDR:
       offset = dissect_isup_initial_address_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_SUBSEQ_ADDR:
       offset = dissect_isup_subsequent_address_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_INFO_REQ:
       offset = dissect_isup_information_request_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_INFO:
       offset = dissect_isup_information_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_CONTINUITY:
       offset = dissect_isup_continuity_message(offset, len);
      break;
    case MESSAGE_TYPE_ADDR_CMPL:
       offset = dissect_isup_address_complete_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_CONNECT:
       offset = dissect_isup_connect_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_FORW_TRANS:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_ANSWER:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_RELEASE:
       offset = dissect_isup_release_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_SUSPEND:
       offset = dissect_isup_suspend_resume_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_RESUME:
       offset = dissect_isup_suspend_resume_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_REL_CMPL:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_CONT_CHECK_REQ:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_RESET_CIRCUIT:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_BLOCKING:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_UNBLOCKING:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_BLOCK_ACK:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_UNBLOCK_ACK:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_CIRC_GRP_RST:
       offset = dissect_isup_circuit_group_reset_query_message(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_BLCK:
       offset = dissect_isup_circuit_group_blocking_messages(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_UNBL:
       offset = dissect_isup_circuit_group_blocking_messages(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_BL_ACK:
       offset = dissect_isup_circuit_group_blocking_messages(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_UNBL_ACK:
       offset = dissect_isup_circuit_group_blocking_messages(offset, len);
      break;
    case MESSAGE_TYPE_FACILITY_REQ:
       offset = dissect_isup_facility_request_accepted_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_FACILITY_ACC:
       offset = dissect_isup_facility_request_accepted_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_FACILITY_REJ:
       offset = dissect_isup_facility_reject_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_LOOP_BACK_ACK:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_PASS_ALONG:
      /* call dissect_isup_message recursively */
//      {	guint8 pa_message_type;
//	pa_message_type = tvb_get_guint8(parameter_tvb, 0);
//	pass_along_item = proto_tree_add_text(isup_tree, parameter_tvb, offset, -1, "Pass-along: %s Message (%u)", val_to_str(pa_message_type, isup_message_type_value_acro, "reserved"), pa_message_type);
//	pass_along_tree = proto_item_add_subtree(pass_along_item, ett_isup_pass_along_message);
//	dissect_isup_message(parameter_tvb, pinfo, pass_along_tree);
	break;
//      }
    case MESSAGE_TYPE_CIRC_GRP_RST_ACK:
       offset = dissect_isup_circuit_group_reset_acknowledgement_message(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_QRY:
       offset = dissect_isup_circuit_group_reset_query_message(offset, len);
      break;
    case MESSAGE_TYPE_CIRC_GRP_QRY_RSP:
       offset = dissect_isup_circuit_group_query_response_message(offset, len);
      break;
    case MESSAGE_TYPE_CALL_PROGRSS:
       offset = dissect_isup_call_progress_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_USER2USER_INFO:
      offset = dissect_isup_user_to_user_information_message(offset, len);
      opt_part_possible =1;
      break;
    case MESSAGE_TYPE_UNEQUIPPED_CIC:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_CONFUSION:
      offset = dissect_isup_confusion_message(offset, len);
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_OVERLOAD:
      /* no dissector necessary since no mandatory parameters included */
      break;
    case MESSAGE_TYPE_CHARGE_INFO:
      /* do nothing since format is a national matter */
		proto_tree_add_text(1, "Format is a national matter: 0x%s", tvb_bytes_to_str(offset, len ));
      break;
    case MESSAGE_TYPE_NETW_RESRC_MGMT:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_FACILITY:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_USER_PART_TEST:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_USER_PART_AVAIL:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_IDENT_REQ:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_IDENT_RSP:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_SEGMENTATION:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    case MESSAGE_TYPE_LOOP_PREVENTION:
      /* no dissector necessary since no mandatory parameters included */
       opt_part_possible =1;
      break;
    default:
		proto_tree_add_text(1, "Unknown Message type (possibly reserved/used in former ISUP version): 0x%s", tvb_bytes_to_str(offset, len ));
      break;
  }

   /* extract pointer to start of optional part (if any) */
   if (opt_part_possible == 1)
   {
     opt_parameter_pointer = tvb_get_guint8(offset);
     if (opt_parameter_pointer > 0)
	 {
       offset += opt_parameter_pointer;
       dissect_isup_optional_parameter(offset, len);
     }
   }

}

/* ------------------------------------------------------------------ */
static void
dissect_isup(int len)
{
	long cic;
	int message_type;

/* Extract message type field */
	message_type = tvb_get_guint8(CIC_OFFSET + CIC_LENGTH);
	cic = tvb_get_guint16i(CIC_OFFSET) & 0x0FFF; /*since upper 4 bits spare */
	
	static char str_title[128];
	sprintf(&str_title[0], "ISUP %s  %s   CIC: %d", val_to_str(message_type, isup_message_type_value_acro, "reserved"),
				val_to_str(message_type, isup_message_type_value, "reserved"),
                cic);
	proto_tree_add_text(0, "%s", &str_title[0]);

		/* dissect CIC in main dissector since pass-along message type carrying complete IUSP message w/o CIC needs
		   recursive message dissector call */
		
		dissect_isup_message(len);
}


int main(int argc, char** argv)
{
	if (argc != 2)
	{
		printf("Usage: %s <ISUP DUMP>\n", argv[0]);
		return 0;
	}
	using namespace std;
	size_t lenStr;
	lenStr = char_traits<char>::length (argv[1]);
	data = argv[1];
	dissect_isup(lenStr/2);

	return 0;
}
